from collections import defaultdict
from typing import Optional

import rich.markup
import rich.tree
import yaml

import autoeft
import autoeft.io.basis as io_basis
from autoeft import exceptions
from autoeft import utils

print = autoeft.print  # noqa: A001


class Stats(defaultdict):
    hc_flag: bool = True

    def hc_mod(self, invariant):
        if self.hc_flag:
            hc_mod = 1 if invariant.is_real() else 2
            return hc_mod if invariant.is_normal() else 0
        return 1

    @property
    def n_families(self):
        return sum(self.hc_mod(family) for family in self)

    @property
    def n_types(self):
        return sum(
            sum(self.hc_mod(op_type) for op_type in types) for types in self.values()
        )

    @property
    def n_terms(self):
        return sum(
            sum(self.hc_mod(op_type) * t for op_type, (t, _) in types.items())
            for types in self.values()
        )

    @property
    def n_operators(self):
        return sum(
            sum(self.hc_mod(op_type) * o for op_type, (_, o) in types.items())
            for types in self.values()
        )

    def yaml(self):
        return {
            str(family): {
                str(op_type): utils.FlowSeq(
                    (self.hc_mod(op_type) * t, self.hc_mod(op_type) * o),
                )
                for op_type, (t, o) in types.items()
            }
            for family, types in self.items()
        }


def count_classes(
    basis,
    select: Optional[list] = None,
    ignore: Optional[list] = None,
):
    classes = Stats(dict)
    for op_info in basis.values():
        op_type = op_info.op_type
        family = op_type.family
        if select and not utils.type_matches(op_type, select):
            continue
        if ignore and utils.type_matches(op_type, ignore):
            continue
        if op_info.n_terms and op_info.n_operators:
            classes[family][op_type] = (op_info.n_terms, op_info.n_operators)
    return classes


def dry_run(
    basis,
    select: Optional[list] = None,
    ignore: Optional[list] = None,
):
    for op_info in basis.values():
        op_type = op_info.op_type
        if select and not utils.type_matches(op_type, select):
            continue
        if ignore and utils.type_matches(op_type, ignore):
            continue
        yield op_type


def main(args) -> int:
    basis_path = args.basis
    basis_file = io_basis.BasisFile(basis_path)
    try:
        model = basis_file.get_model()
        basis = basis_file.get_basis()
        dimension = basis_file.get_extra().dimension
        hc_flag = not args.no_hc
    except exceptions.BasisNotFoundError as e:
        errmsg = f"{e}\nInvalid basis directory '{basis_file.basis_path}'."
        print(errmsg)
        return 1
    if args.verbose:
        print(model)
        print()
    try:
        select = (
            [yaml.safe_load(s.replace(":", ": ")) for s in args.select]
            if args.select
            else None
        )
        ignore = (
            [yaml.safe_load(i.replace(":", ": ")) for i in args.ignore]
            if args.ignore
            else None
        )
    except yaml.YAMLError as e:
        errmsg = f"{e}\nCould not parse select/ignore options."
        print(errmsg)
        return 1
    if args.dry_run:
        print(
            f"Dry run for "
            f"[b]{rich.markup.escape(model.name)}[/b] @ [i]d={dimension}[/i]",
        )
        print(f"[i]select[/i]: {select}")
        print(f"[i]ignore[/i]: {ignore}")
        print()
        print("[b]types:[/b]")
        for op_type in dry_run(
            basis,
            select,
            ignore,
        ):
            print(f"[bold white]{op_type}[/bold white]", dest="terminal")
            print(op_type, dest="log")
        return 0

    print(
        f"Counting invariant classes for "
        f"[b]{rich.markup.escape(model.name)}[/b] @ [i]d={dimension}[/i]",
    )
    if not hc_flag:
        print("[i](no Hermitian conjugation implied)[/i]")
    stats = count_classes(basis, select, ignore)
    stats.hc_flag = hc_flag
    if args.verbose:
        eft_tree = rich.tree.Tree(
            f"[d]{rich.markup.escape(model.name)}[/d] @ [i]d={dimension}[/i]",
            style="bold",
            highlight=True,
        )
        for family, types in stats.items():
            if hc_flag and not family.is_normal():
                continue
            hc_str = (
                "[dim italic not bold] + h.c.[/]"
                if hc_flag and family.is_complex()
                else ""
            )
            family_tree = eft_tree.add(
                str(family) + hc_str,
                style="bold",
                highlight=True,
            )
            for op_type, (n_term, n_operator) in types.items():
                if hc_flag and not op_type.is_normal():
                    continue
                hc_mod = 2 if hc_flag and op_type.is_complex() else 1
                hc_str = (
                    "[dim italic not bold] + h.c.[/]"
                    if hc_flag and op_type.is_complex()
                    else ""
                )
                type_tree = family_tree.add(str(op_type) + hc_str)
                type_tree.add(f"#terms={hc_mod*n_term}", style="not bold")
                type_tree.add(f"#operators={hc_mod*n_operator}", style="not bold")
        autoeft.console.rule()
        print(eft_tree)
        autoeft.console.rule()
    print(f"[b]#[/b]families={stats.n_families:,}")
    print(f"[b]#[/b]types={stats.n_types:,}")
    print(f"[b]#[/b]terms={stats.n_terms:,}")
    print(f"[b]#[/b]operators={stats.n_operators:,}")
    if args.output:
        stats_file = args.output.resolve()
        data = {}
        if args.verbose:
            data = {"classes": stats.yaml()}
        data["n_families"] = stats.n_families
        data["n_types"] = stats.n_types
        data["n_terms"] = stats.n_terms
        data["n_operators"] = stats.n_operators
        try:
            hc_comment = "" if hc_flag else "# no Hermitian conjugation implied\n"
            stats_file.write_text(
                hc_comment + yaml.dump(data, sort_keys=False, Dumper=utils.YamlDumper),
            )
        except IsADirectoryError:
            errmsg = f"'{stats_file}' is not a valid output directory."
            print(errmsg)
            return 1
    print(f"Saved numbers in {stats_file}")
    return 0



import itertools
from collections import Counter
from collections import defaultdict
from collections import deque
from math import prod
from typing import Counter as Count
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import pydantic
import rich.console
import rich.layout
import rich.live
import rich.markup
import rich.progress
import rich.rule
import rich.tree
import sage.matrix.constructor
import semver
import yaml
from sage.rings.integer_ring import ZZ

import autoeft
import autoeft.base.basis as eft_basis
import autoeft.base.classifications as classes
import autoeft.base.model as ir_model
import autoeft.construction.permutation_symmetries as sym_const
import autoeft.construction.symmetrizers as sym_reprs
import autoeft.form
import autoeft.io.basis as io_basis
from autoeft import exceptions
from autoeft import utils
from autoeft.base import tensors
from autoeft.combinatorics import generators
from autoeft.construction import invariants

print = autoeft.print  # noqa: A001


def construct_permutation_matrix(
    model: ir_model.Model,
    op_type: classes.Type,
    lorentz_tableaux: List[tensors.LorentzTableau],
    sun_groups_tableaux: Dict[str, List[tensors.SUNTableau]],
    combination: Count[Tuple[sym_reprs.GroupSymmetrizer, ...]],
    symmetrizer_matrices: Dict[
        sym_reprs.GroupSymmetrizer,
        Tuple[sage.matrix.constructor.Matrix, sage.matrix.constructor.Matrix],
    ],
    inverse_gram_matrices: Dict[str, sage.matrix.constructor.Matrix],
    threads: Optional[int] = None,
) -> Tuple[sage.matrix.constructor.Matrix, sage.matrix.constructor.Matrix]:
    ranks = defaultdict(set)
    pivots = defaultdict(set)
    permutation_matrices = []
    reduced_permutation_matrices = []
    for symmetrizer, cgc in combination.items():
        decomposition_matrix = sage.matrix.constructor.matrix.identity(1)
        reduced_decomposition_matrix = sage.matrix.constructor.matrix.identity(1)
        for group_symmetrizer in symmetrizer:
            if group_symmetrizer in symmetrizer_matrices:
                (
                    symmetrizer_matrix,
                    reduced_matrix,
                ) = symmetrizer_matrices[group_symmetrizer]
            else:
                group_name = group_symmetrizer.group_name
                if group_name == model.symmetries.lorentz_group.name:
                    basis = lorentz_tableaux
                    symmetrizer_matrix = sym_reprs.construct_lorentz_matrix(
                        op_type,
                        group_symmetrizer,
                        basis,
                    )
                else:
                    basis = sun_groups_tableaux[group_name]
                    group = model.symmetries.sun_groups[group_name]
                    symmetrizer_matrix = sym_reprs.construct_sun_matrix(
                        op_type,
                        group_symmetrizer,
                        basis,
                        group,
                        inverse_gram_matrices,
                        threads,
                    )
                rank = symmetrizer_matrix.rank()
                pivot = symmetrizer_matrix.pivot_rows()
                reduced_matrix = (
                    symmetrizer_matrix[pivot, :]
                    if pivot
                    else sage.matrix.constructor.matrix()
                )
                symmetrizer_matrices[group_symmetrizer] = (
                    symmetrizer_matrix,
                    reduced_matrix,
                )
                ranks[group_name].add(rank)
                pivots[group_name].add(pivot)
            decomposition_matrix = decomposition_matrix.tensor_product(
                symmetrizer_matrix,
                subdivide=False,
            )
            reduced_decomposition_matrix = reduced_decomposition_matrix.tensor_product(
                reduced_matrix,
                subdivide=False,
            )
        permutation_matrices.append(cgc * decomposition_matrix)
        reduced_permutation_matrices.append(cgc * reduced_decomposition_matrix)
    assert all(len(ranks) == 1 for ranks in ranks.values())
    assert all(len(pivots) == 1 for pivots in pivots.values())
    permutation_matrix = sum(permutation_matrices)
    reduced_permutation_matrix = sum(reduced_permutation_matrices)
    if reduced_permutation_matrix:
        reduced_permutation_matrix *= reduced_permutation_matrix.denominator()
        reduced_permutation_matrix = reduced_permutation_matrix.change_ring(ZZ)
        reduced_permutation_matrix /= reduced_permutation_matrix.gcd()
    reduced_permutation_matrix = reduced_permutation_matrix.change_ring(ZZ)
    return permutation_matrix, reduced_permutation_matrix


def plain_main(
    model: ir_model.Model,
    dimension: int,
    hc_flag: bool,
    gr_flag: bool,
    basis_file: Optional[io_basis.BasisFile] = None,
    select: Optional[list] = None,
    ignore: Optional[list] = None,
    overwrite: bool = False,
    threads: Optional[int] = None,
) -> Tuple[eft_basis.Basis, Tuple[Counter, Counter, Counter, Counter]]:
    basis = eft_basis.Basis(model)
    ct_family, ct_type, ct_term, ct_operator = (
        Counter(),
        Counter(),
        Counter(),
        Counter(),
    )
    tct_family, tct_type, tct_term, tct_operator = (
        Counter(),
        Counter(),
        Counter(),
        Counter(),
    )
    operator_status = rich.progress.Progress(
        rich.progress.SpinnerColumn(spinner_name="simpleDotsScrolling"),
        rich.progress.TextColumn("{task.description}", style="bold"),
    )
    status_task = operator_status.add_task("")
    stats_progress = rich.progress.Progress(
        rich.progress.TextColumn(
            "#[repr.attrib_name]{task.description}[/repr.attrib_name]"
            "="
            "[repr.number]{task.total:,}[/repr.number]"
            " ([repr.number]{task.fields[exist]:,}[/repr.number] existing,"
            " [repr.number]{task.completed:,}[/repr.number] constructed in this run)",
        ),
    )
    families_task = stats_progress.add_task("families", total=0, exist=0)
    types_task = stats_progress.add_task("types", total=0, exist=0)
    terms_task = stats_progress.add_task("terms", total=0, exist=0)
    operators_task = stats_progress.add_task("operators", total=0, exist=0)
    progress_group = rich.console.Group(
        operator_status,
        stats_progress,
    )
    with rich.live.Live(
        progress_group,
        console=autoeft.console,
        transient=True,
    ) as live:
        for family in invariants.construct_families(dimension, hc_flag, gr_flag):
            hc_mod_f = 1 if hc_flag or family.is_real() else 2
            for op_type in invariants.construct_types(family, model, hc_flag):
                hc_mod_t = 1 if hc_flag or op_type.is_real() else 2
                operator_status.update(status_task, description=str(op_type))
                live.refresh()
                if basis_file:
                    op_stats = basis_file.get_operator_raw_numbers(op_type.full_path())
                    if op_stats:
                        terms_per_type, operators_per_type = op_stats
                        if terms_per_type:
                            tct_family[family] = hc_mod_f
                            tct_type[op_type] = hc_mod_t
                            tct_term[op_type] = hc_mod_t * terms_per_type
                            tct_operator[op_type] = hc_mod_t * operators_per_type
                            stats_progress.update(
                                families_task,
                                total=sum(tct_family.values()),
                                exist=sum(tct_family.values())
                                - sum(ct_family.values()),
                            )
                            stats_progress.update(
                                types_task,
                                total=sum(tct_type.values()),
                                exist=sum(tct_type.values()) - sum(ct_type.values()),
                            )
                            stats_progress.update(
                                terms_task,
                                total=sum(tct_term.values()),
                                exist=sum(tct_term.values()) - sum(ct_term.values()),
                            )
                            stats_progress.update(
                                operators_task,
                                total=sum(tct_operator.values()),
                                exist=sum(tct_operator.values())
                                - sum(ct_operator.values()),
                            )
                        if not overwrite:
                            continue
                if select and not utils.type_matches(op_type, select):
                    continue
                if ignore and utils.type_matches(op_type, ignore):
                    continue
                try:
                    lorentz_tableaux = invariants.construct_lorentz_tableaux(family)
                    sun_groups_tableaux = {
                        group_name: invariants.construct_sun_tableaux(op_type, group)
                        for group_name, group in model.symmetries.sun_groups.items()
                    }
                except exceptions.NoContractionException:
                    continue  # no invariant contractions found
                terms_per_type = 0
                operators_per_type = 0
                permutations_info = []
                symmetrizer_matrices = {}  # cache symmetrizer matrices
                inverse_gram_matrices = {}  # cache inverse Gram matrices
                allowed_group_permutations = sym_const.construct_group_permutations(
                    op_type,
                    model.symmetries,
                )
                permutation_symmetries = sym_const.construct_permutation_symmetries(
                    op_type,
                    allowed_group_permutations,
                )
                for permutation_symmetry in permutation_symmetries:
                    factorized_symmetries = sym_const.construct_factorized_symmetries(
                        *permutation_symmetry,
                    )
                    for symmetry, combination in factorized_symmetries:
                        (
                            permutation_matrix,
                            reduced_permutation_matrix,
                        ) = construct_permutation_matrix(
                            model,
                            op_type,
                            lorentz_tableaux,
                            sun_groups_tableaux,
                            combination,
                            symmetrizer_matrices,
                            inverse_gram_matrices,
                            threads,
                        )
                        terms_per_symmetry = reduced_permutation_matrix.nrows()
                        operators_per_symmetry = terms_per_symmetry * prod(
                            permutation.dimension_SUN(
                                model.fields[field_name].generations,
                            )
                            for field_name, permutation in symmetry.items()
                        )
                        if terms_per_symmetry:
                            permutation_info = (
                                eft_basis.PermutationSymmetryInfo.construct(
                                    symmetry=symmetry,
                                    n_terms=terms_per_symmetry,
                                    n_operators=operators_per_symmetry,
                                    matrix=reduced_permutation_matrix,
                                )
                            )
                            permutations_info.append(permutation_info)
                            terms_per_type += terms_per_symmetry
                            operators_per_type += operators_per_symmetry
                if terms_per_type:
                    type_info = (
                        op_type.to_counter(),
                        "real" if op_type.is_real() else "complex",
                    )
                    generations_info = Counter(
                        {f.name: f.generations for f, _ in op_type},
                    )
                    invariants_info = {
                        model.symmetries.lorentz_group.name: list(
                            zip(lorentz_tableaux, itertools.repeat(1)),
                        ),
                    }
                    invariants_info.update(
                        {
                            sun_group: list(zip(sun_tableaux, itertools.repeat(1)))
                            for sun_group, sun_tableaux in sun_groups_tableaux.items()
                        },
                    )
                    operator_info = eft_basis.OperatorInfoPermutation.construct(
                        model=model,
                        type=type_info,
                        generations=generations_info,
                        n_terms=terms_per_type,
                        n_operators=operators_per_type,
                        invariants=invariants_info,
                        permutation_symmetries=permutations_info,
                    )
                    basis[repr(operator_info)] = operator_info
                    if basis_file:
                        basis_file.add_operator_info(operator_info)
                    print(repr(op_type), dest="log")
                    ct_family[family] = hc_mod_f
                    ct_type[op_type] = hc_mod_t
                    ct_term[op_type] = hc_mod_t * terms_per_type
                    ct_operator[op_type] = hc_mod_t * operators_per_type
                    tct_family[family] = hc_mod_f
                    tct_type[op_type] = hc_mod_t
                    tct_term[op_type] = hc_mod_t * terms_per_type
                    tct_operator[op_type] = hc_mod_t * operators_per_type
                    stats_progress.update(
                        families_task,
                        total=sum(tct_family.values()),
                        completed=sum(ct_family.values()),
                        exist=sum(tct_family.values()) - sum(ct_family.values()),
                    )
                    stats_progress.update(
                        types_task,
                        total=sum(tct_type.values()),
                        advance=ct_type[op_type],
                        exist=sum(tct_type.values()) - sum(ct_type.values()),
                    )
                    stats_progress.update(
                        terms_task,
                        total=sum(tct_term.values()),
                        advance=ct_term[op_type],
                        exist=sum(tct_term.values()) - sum(ct_term.values()),
                    )
                    stats_progress.update(
                        operators_task,
                        total=sum(tct_operator.values()),
                        advance=ct_operator[op_type],
                        exist=sum(tct_operator.values()) - sum(ct_operator.values()),
                    )
        operator_status.update(status_task, visible=False)
    fam_total, fam_constructed = sum(tct_family.values()), sum(ct_family.values())
    type_total, type_constructed = sum(tct_type.values()), sum(ct_type.values())
    term_total, term_constructed = sum(tct_term.values()), sum(ct_term.values())
    op_total, op_constructed = sum(tct_operator.values()), sum(ct_operator.values())
    print(
        f"[b]#[/b]families=[repr.number]{fam_total:,}[/repr.number]"
        f" ([repr.number]{fam_total-fam_constructed:,}[/repr.number] existing,"
        f" [repr.number]{fam_constructed:,}[/repr.number] constructed in this run)",
    )
    print(
        f"[b]#[/b]types=[repr.number]{type_total:,}[/repr.number]"
        f" ([repr.number]{type_total-type_constructed:,}[/repr.number] existing,"
        f" [repr.number]{type_constructed:,}[/repr.number] constructed in this run)",
    )
    print(
        f"[b]#[/b]terms=[repr.number]{term_total:,}[/repr.number]"
        f" ([repr.number]{term_total-term_constructed:,}[/repr.number] existing,"
        f" [repr.number]{term_constructed:,}[/repr.number] constructed in this run)",
    )
    print(
        f"[b]#[/b]operators=[repr.number]{op_total:,}[/repr.number]"
        f" ([repr.number]{op_total-op_constructed:,}[/repr.number] existing,"
        f" [repr.number]{op_constructed:,}[/repr.number] constructed in this run)",
    )
    if basis_file and op_constructed:
        print(f"Saved operators in {basis_file.basis_path}")
    return basis, (tct_family, tct_type, tct_term, tct_operator)


def verbose_main(
    model: ir_model.Model,
    dimension: int,
    hc_flag: bool,
    gr_flag: bool,
    basis_file: Optional[io_basis.BasisFile] = None,
    select: Optional[list] = None,
    ignore: Optional[list] = None,
    overwrite: bool = False,
    threads: Optional[int] = None,
) -> Tuple[eft_basis.Basis, Tuple[Counter, Counter, Counter, Counter]]:
    basis = eft_basis.Basis(model)
    ct_family, ct_type, ct_term, ct_operator = (
        Counter(),
        Counter(),
        Counter(),
        Counter(),
    )
    tct_family, tct_type, tct_term, tct_operator = (
        Counter(),
        Counter(),
        Counter(),
        Counter(),
    )
    eft_tree = rich.tree.Tree(
        f"[d]{rich.markup.escape(model.name)}[/d] @ [i]d={dimension}[/i]",
        style="bold",
        highlight=True,
    )
    render_eft_tree = rich.tree.Tree(
        f"[d]{rich.markup.escape(model.name)}[/d] @ [i]d={dimension}[/i]",
        style="bold",
        highlight=True,
    )
    render_eft_tree.children = deque(maxlen=1)
    operator_status = rich.progress.Progress(
        rich.progress.SpinnerColumn(spinner_name="simpleDotsScrolling"),
        rich.progress.TextColumn("{task.description}", style="bold"),
    )
    status_task = operator_status.add_task("")
    layout = rich.layout.Layout(name="root")
    layout.split_column(
        rich.layout.Layout(rich.rule.Rule(), name="header", size=1),
        rich.layout.Layout(render_eft_tree, name="tree"),
        rich.layout.Layout(rich.rule.Rule(), name="rule", size=1),
        rich.layout.Layout(operator_status, name="status", size=1),
    )
    tree_height = layout.render(autoeft.console, autoeft.console.options)[
        layout["tree"]
    ].region.height
    with rich.live.Live(layout, console=autoeft.console, transient=True) as live:
        for family in invariants.construct_families(dimension, hc_flag, gr_flag):
            layout["header"].update(rich.rule.Rule(family.tuple_repr()))
            hc_mod_f = 1 if hc_flag or family.is_real() else 2
            hc_str = (
                "" if hc_flag or family.is_real() else "[dim italic not bold] + h.c.[/]"
            )
            family_tree = eft_tree.add(
                str(family) + hc_str,
                style="bold",
                highlight=True,
            )
            render_family_tree = render_eft_tree.add(
                str(family) + hc_str,
                style="bold",
                highlight=True,
            )
            render_family_tree.children = deque(maxlen=(tree_height - 2) // 3)

            for op_type in invariants.construct_types(family, model, hc_flag):
                hc_mod_t = 1 if hc_flag or op_type.is_real() else 2
                hc_str = (
                    ""
                    if hc_flag or op_type.is_real()
                    else "[dim italic not bold] + h.c.[/]"
                )
                operator_status.update(status_task, description=str(op_type))
                live.refresh()
                if basis_file:
                    op_stats = basis_file.get_operator_raw_numbers(op_type.full_path())
                    if op_stats:
                        terms_per_type, operators_per_type = op_stats
                        if terms_per_type:
                            tct_family[family] = hc_mod_f
                            tct_type[op_type] = hc_mod_t
                            tct_term[op_type] = hc_mod_t * terms_per_type
                            tct_operator[op_type] = hc_mod_t * operators_per_type
                        if not overwrite:
                            if tct_term[op_type]:
                                type_tree = family_tree.add(str(op_type) + hc_str)
                                type_tree.add(
                                    f"#terms={tct_term[op_type]}",
                                    style="not bold dim",
                                )
                                type_tree.add(
                                    f"#operators={tct_operator[op_type]}",
                                    style="not bold dim",
                                )
                                render_type_tree = render_family_tree.add(
                                    str(op_type) + hc_str,
                                )
                                render_type_tree.add(
                                    f"#terms={tct_term[op_type]}",
                                    style="not bold dim",
                                )
                                render_type_tree.add(
                                    f"#operators={tct_operator[op_type]}",
                                    style="not bold dim",
                                )
                            continue
                if select and not utils.type_matches(op_type, select):
                    if tct_term[op_type]:
                        type_tree = family_tree.add(str(op_type) + hc_str)
                        type_tree.add(
                            f"#terms={tct_term[op_type]}",
                            style="not bold dim",
                        )
                        type_tree.add(
                            f"#operators={tct_operator[op_type]}",
                            style="not bold dim",
                        )
                        render_type_tree = render_family_tree.add(str(op_type) + hc_str)
                        render_type_tree.add(
                            f"#terms={tct_term[op_type]}",
                            style="not bold dim",
                        )
                        render_type_tree.add(
                            f"#operators={tct_operator[op_type]}",
                            style="not bold dim",
                        )
                    continue
                if ignore and utils.type_matches(op_type, ignore):
                    if tct_term[op_type]:
                        type_tree = family_tree.add(str(op_type) + hc_str)
                        type_tree.add(
                            f"#terms={tct_term[op_type]}",
                            style="not bold dim",
                        )
                        type_tree.add(
                            f"#operators={tct_operator[op_type]}",
                            style="not bold dim",
                        )
                        render_type_tree = render_family_tree.add(str(op_type) + hc_str)
                        render_type_tree.add(
                            f"#terms={tct_term[op_type]}",
                            style="not bold dim",
                        )
                        render_type_tree.add(
                            f"#operators={tct_operator[op_type]}",
                            style="not bold dim",
                        )
                    continue
                try:
                    lorentz_tableaux = invariants.construct_lorentz_tableaux(family)
                    sun_groups_tableaux = {
                        group_name: invariants.construct_sun_tableaux(op_type, group)
                        for group_name, group in model.symmetries.sun_groups.items()
                    }
                except exceptions.NoContractionException:
                    continue  # no invariant contractions found
                terms_per_type = 0
                operators_per_type = 0
                permutations_info = []
                symmetrizer_matrices = {}  # cache symmetrizer matrices
                inverse_gram_matrices = {}  # cache inverse Gram matrices
                allowed_group_permutations = sym_const.construct_group_permutations(
                    op_type,
                    model.symmetries,
                )
                permutation_symmetries = sym_const.construct_permutation_symmetries(
                    op_type,
                    allowed_group_permutations,
                )
                for permutation_symmetry in permutation_symmetries:
                    factorized_symmetries = sym_const.construct_factorized_symmetries(
                        *permutation_symmetry,
                    )
                    for symmetry, combination in factorized_symmetries:
                        (
                            permutation_matrix,
                            reduced_permutation_matrix,
                        ) = construct_permutation_matrix(
                            model,
                            op_type,
                            lorentz_tableaux,
                            sun_groups_tableaux,
                            combination,
                            symmetrizer_matrices,
                            inverse_gram_matrices,
                            threads,
                        )
                        terms_per_symmetry = reduced_permutation_matrix.nrows()
                        operators_per_symmetry = terms_per_symmetry * prod(
                            permutation.dimension_SUN(
                                model.fields[field_name].generations,
                            )
                            for field_name, permutation in symmetry.items()
                        )
                        if terms_per_symmetry:
                            permutation_info = (
                                eft_basis.PermutationSymmetryInfo.construct(
                                    symmetry=symmetry,
                                    n_terms=terms_per_symmetry,
                                    n_operators=operators_per_symmetry,
                                    matrix=reduced_permutation_matrix,
                                )
                            )
                            permutations_info.append(permutation_info)
                            terms_per_type += terms_per_symmetry
                            operators_per_type += operators_per_symmetry
                if terms_per_type:
                    type_info = (
                        op_type.to_counter(),
                        "real" if op_type.is_real() else "complex",
                    )
                    generations_info = Counter(
                        {f.name: f.generations for f, _ in op_type},
                    )
                    invariants_info = {
                        model.symmetries.lorentz_group.name: list(
                            zip(lorentz_tableaux, itertools.repeat(1)),
                        ),
                    }
                    invariants_info.update(
                        {
                            sun_group: list(zip(sun_tableaux, itertools.repeat(1)))
                            for sun_group, sun_tableaux in sun_groups_tableaux.items()
                        },
                    )
                    operator_info = eft_basis.OperatorInfoPermutation.construct(
                        model=model,
                        type=type_info,
                        generations=generations_info,
                        n_terms=terms_per_type,
                        n_operators=operators_per_type,
                        invariants=invariants_info,
                        permutation_symmetries=permutations_info,
                    )
                    basis[repr(operator_info)] = operator_info
                    if basis_file:
                        basis_file.add_operator_info(operator_info)
                    print(repr(op_type), dest="log")
                    ct_family[family] = hc_mod_f
                    ct_type[op_type] = hc_mod_t
                    ct_term[op_type] = hc_mod_t * terms_per_type
                    ct_operator[op_type] = hc_mod_t * operators_per_type
                    tct_family[family] = hc_mod_f
                    tct_type[op_type] = hc_mod_t
                    tct_term[op_type] = hc_mod_t * terms_per_type
                    tct_operator[op_type] = hc_mod_t * operators_per_type
                    type_tree = family_tree.add(str(op_type) + hc_str)
                    type_tree.add(f"#terms={ct_term[op_type]}", style="not bold")
                    type_tree.add(
                        f"#operators={ct_operator[op_type]}",
                        style="not bold",
                    )
                    render_type_tree = render_family_tree.add(str(op_type) + hc_str)
                    render_type_tree.add(f"#terms={ct_term[op_type]}", style="not bold")
                    render_type_tree.add(
                        f"#operators={ct_operator[op_type]}",
                        style="not bold",
                    )
        layout["header"].update(rich.rule.Rule())
    print(eft_tree)
    autoeft.console.rule()
    fam_total, fam_constructed = sum(tct_family.values()), sum(ct_family.values())
    type_total, type_constructed = sum(tct_type.values()), sum(ct_type.values())
    term_total, term_constructed = sum(tct_term.values()), sum(ct_term.values())
    op_total, op_constructed = sum(tct_operator.values()), sum(ct_operator.values())
    print(
        f"[b]#[/b]families=[repr.number]{fam_total:,}[/repr.number]"
        f" ([repr.number]{fam_total-fam_constructed:,}[/repr.number] existing,"
        f" [repr.number]{fam_constructed:,}[/repr.number] constructed in this run)",
    )
    print(
        f"[b]#[/b]types=[repr.number]{type_total:,}[/repr.number]"
        f" ([repr.number]{type_total-type_constructed:,}[/repr.number] existing,"
        f" [repr.number]{type_constructed:,}[/repr.number] constructed in this run)",
    )
    print(
        f"[b]#[/b]terms=[repr.number]{term_total:,}[/repr.number]"
        f" ([repr.number]{term_total-term_constructed:,}[/repr.number] existing,"
        f" [repr.number]{term_constructed:,}[/repr.number] constructed in this run)",
    )
    print(
        f"[b]#[/b]operators=[repr.number]{op_total:,}[/repr.number]"
        f" ([repr.number]{op_total-op_constructed:,}[/repr.number] existing,"
        f" [repr.number]{op_constructed:,}[/repr.number] constructed in this run)",
    )
    if basis_file and op_constructed:
        print(f"Saved operators in {basis_file.basis_path}")
    return basis, (tct_family, tct_type, tct_term, tct_operator)


def dry_run(
    model: ir_model.Model,
    dimension: int,
    hc_flag: bool,
    gr_flag: bool,
    select: Optional[list] = None,
    ignore: Optional[list] = None,
):
    for family in invariants.construct_families(dimension, hc_flag, gr_flag):
        for op_type in invariants.construct_types(family, model, hc_flag):
            if select and not utils.type_matches(op_type, select):
                continue
            if ignore and utils.type_matches(op_type, ignore):
                continue
            yield op_type


def main(args) -> int:
    form_version_min = semver.Version(4, 3)
    if autoeft.form.version < form_version_min:
        errmsg = (
            f"The version of FORM installed is not compatible with {autoeft.program}."
            "\n"
            f"Please install FORM version {form_version_min} or higher."
        )
        raise exceptions.RequirementVersionError(errmsg)
    model_path = args.model.resolve()
    dimension = args.dimension
    output_path = args.output.resolve()
    if args.name:
        output_path /= args.name
    else:
        output_path /= f"{model_path.stem}-eft"
    generators.default_path = args.generators.resolve()
    try:
        with model_path.open("r") as model_file:
            model = ir_model.Model(**yaml.safe_load(model_file))
    except (FileNotFoundError, IsADirectoryError):
        errmsg = f"Model file not found {model_path}."
        print(errmsg)
        return 1
    except pydantic.ValidationError as e:
        errmsg = f"{e}\nCould not validate model."
        print(errmsg)
        return 1
    except yaml.YAMLError as e:
        errmsg = f"{e}\nCould not parse model file."
        print(errmsg)
        return 1
    hc_flag = not args.no_hc
    gr_flag = any(abs(f.helicity) == 2 for f in model.fields.values())
    print(model)
    print()

    try:
        select = (
            [yaml.safe_load(s.replace(":", ": ")) for s in args.select]
            if args.select
            else None
        )
        ignore = (
            [yaml.safe_load(i.replace(":", ": ")) for i in args.ignore]
            if args.ignore
            else None
        )
    except yaml.YAMLError as e:
        errmsg = f"{e}\nCould not parse select/ignore options."
        print(errmsg)
        return 1
    if args.dry_run:
        print(
            f"Dry run for "
            f"[b]{rich.markup.escape(model.name)}[/b] @ [i]d={dimension}[/i]",
        )
        print(f"[i]select[/i]: {select}")
        print(f"[i]ignore[/i]: {ignore}")
        print()
        print("[b]types:[/b]")
        for op_type in dry_run(
            model,
            dimension,
            hc_flag,
            gr_flag,
            select,
            ignore,
        ):
            print(f"[bold white]{op_type}[/bold white]", dest="terminal")
            print(op_type, dest="log")
        return 0

    if args.threads <= 0:
        errmsg = f"Number of threads must be greater than 0, not '{args.threads}'."
        print(errmsg)
        return 1

    wd = output_path / ascii(dimension)
    wd.mkdir(parents=True, exist_ok=True)
    basis_path = wd / "basis"
    basis_file = io_basis.BasisFile(basis_path)
    if basis_file:
        extra = basis_file.get_extra()
        print(f"Existing operators found in {basis_path}")
        print(f"Last run at {extra.timestamp}")
        existing_model = basis_file.get_model()
        if model != existing_model:
            errmsg = (
                "Incompatible model files."
                "\n"
                f"The basis directory {basis_path} already exists"
                " but it was created using a different model."
                "\n"
                "Please delete the existing basis directory"
                " or provide a different output path using"
                " the --name or --output option."
            )
            print(errmsg)
            return 1
        extra_diff = {
            k: (v, vars(args)[k])
            for k, v in extra.args.items()
            if k in {"select", "ignore", "no_hc"} and v != vars(args)[k]
        }
        if extra_diff:
            erropt = "\n".join(
                f" - {option}: {prev} -> {curr}"
                for option, (prev, curr) in extra_diff.items()
            )
            errmsg = (
                "Incompatible options:"
                f"\n{erropt}\n"
                f"The basis directory {basis_path} already exists"
                " but it was created using different options."
                "\n"
                "Please delete the existing basis directory"
                " or provide a different output path using"
                " the --name or --output option."
            )
            print(errmsg)
            return 1
        print()
    else:
        basis_file.set_model(model)
    basis_file.set_extra(output_path.name, dimension, args)

    print(
        f"Constructing operator basis for "
        f"[b]{rich.markup.escape(model.name)}[/b] @ [i]d={dimension}[/i]",
    )
    if not hc_flag:
        print("(Hermitian conjugation implied)")
    main_func = verbose_main if not args.quiet and args.verbose else plain_main
    basis, (ct_family, ct_type, ct_term, ct_operator) = main_func(
        model,
        dimension,
        hc_flag,
        gr_flag,
        basis_file,
        select,
        ignore,
        args.overwrite,
        args.threads,
    )
    stats = {}
    stats["n_families"] = int(sum(ct_family.values()))
    stats["n_types"] = int(sum(ct_type.values()))
    stats["n_terms"] = int(sum(ct_term.values()))
    stats["n_operators"] = int(sum(ct_operator.values()))
    stats_path = wd / "stats.yml"
    stats_path.write_text(yaml.safe_dump(stats, sort_keys=False))
    return 0



import itertools
import math
import string
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Counter
from typing import Dict
from typing import List
from typing import Literal
from typing import Tuple

import pydantic
import yaml

import autoeft
import autoeft.base.classifications as classes
import autoeft.base.model as ir_model
from autoeft import utils
from autoeft.base import tensors
from autoeft.combinatorics import young
from autoeft.tex import tex


class OperatorInfo(pydantic.BaseModel):
    model: ir_model.Model
    version: utils.Version = autoeft.version
    type: Tuple[Counter[str], Literal["real", "complex"]]
    generations: Counter[str]
    n_terms: pydantic.PositiveInt
    n_operators: pydantic.PositiveInt
    invariants: Dict[str, List[Tuple[young.Tableau, int]]]

    def __str__(self):
        return str(self.op_type)

    def __repr__(self):
        return repr(self.op_type)

    def __tex__(self):
        invariants_content = {}
        for group_name, invariants in self.invariants.items():
            group = self.model.symmetries[group_name]
            group_tex = tex(group)
            equations = "\\\\\n".join(
                rf"\mathcal{{O}}_{{{i}}}^{{{group_tex}}} = "
                + invariant[0].render_tableau_tex(self.op_type, group, invariant[1])
                for i, invariant in enumerate(invariants, start=1)
            )
            equation = f"\\begin{{gather}}\n{equations}\n\\end{{gather}}"
            invariants_content[f"${group_tex}$"] = f"\n{equation}"
        invariants_items = (
            rf"\item[{item}]{value}" for item, value in invariants_content.items()
        )
        content = {
            r"$\bullet$": (
                rf"$\mathcal{{N}}_{{\text{{terms}}}}={self.n_terms}"
                r"$\quad\&\quad"
                rf"$\mathcal{{N}}_{{\text{{operators}}}}={self.n_operators}$"
                f"\n\\hfill{{\\small type: \\textit{{{self.type[1]}}}}}"
            ),
            "invariants": "\\hfill\n"
            + "\n".join(
                (
                    r"\begin{description}",
                    *invariants_items,
                    r"\end{description}",
                ),
            ),
        }
        items = (rf"\item[{item}]{value}" for item, value in content.items())
        return "\n".join((r"\begin{description}", *items, r"\end{description}"))

    def __eq__(self, other) -> bool:
        if self.version.to_tuple() != other.version.to_tuple():
            return False
        if self.invariants != other.invariants:
            return False
        return self.dict(exclude={"version", "invariants"}) == other.dict(
            exclude={"version", "invariants"},
        )

    @pydantic.validator("version")
    def parse_version(cls, v):
        if not v.is_compatible(autoeft.version):
            errmsg = (
                f"The version {v} of the operator file is not compatible with"
                f" {autoeft.program} {autoeft.version}."
                "\n"
                f"Please make sure you have the latest version of {autoeft.program}"
                ", or obtain a compatible version of the file."
            )
            raise ValueError(errmsg)
        return v

    @pydantic.validator("invariants", pre=True)
    def parse_invariants(cls, v, values):
        invariants = defaultdict(list)
        if "model" in values:
            model = values["model"]
            lorentz_group = model.symmetries.lorentz_group
            for group_name, contractions in v.items():
                for contraction in contractions.values():
                    T = (
                        tensors.LorentzTableau
                        if group_name == lorentz_group.name
                        else tensors.SUNTableau
                    )
                    tab, sign = T.parse_tableau(contraction)
                    invariants[group_name].append((tab, sign))
        return invariants

    @property
    def op_type(self):
        return classes.Type.from_counter(self.type[0], self.model)

    @property
    def nD(self) -> int:
        return self.op_type.nD

    @property
    def d(self) -> int:
        return self.op_type.d

    @property
    def N(self):
        return self.op_type.family.N

    @property
    def nl(self):
        return self.op_type.family.nl

    @property
    def nr(self):
        return self.op_type.family.nr

    def is_real(self) -> bool:
        return self.op_type.is_real()

    def is_complex(self) -> bool:
        return self.op_type.is_complex()

    def path(self) -> Path:
        return self.op_type.full_path()

    def yaml_dict(self):
        operator_data = self.dict(
            exclude={"model", "invariants", "permutation_symmetries", "expanded"},
        )
        invariants_data = defaultdict(utils.BlockMap)
        for group_name, tableaux in self.invariants.items():
            for i, (tableau, sign) in enumerate(tableaux, start=1):
                if group_name == self.model.symmetries.lorentz_group.name:
                    invariants_data[group_name][f"O({group_name},{i})"] = (
                        tableau.render_tableau(self.op_type, sign)
                    )
                else:
                    invariants_data[group_name][f"O({group_name},{i})"] = (
                        tableau.render_tableau(self.op_type, group_name, sign)
                    )
        operator_data["invariants"] = invariants_data
        return operator_data

    def yaml(self):
        return yaml.dump(
            self.yaml_dict(),
            Dumper=utils.YamlDumper,
            sort_keys=False,
            default_flow_style=None,
            width=math.inf,
        )


class PermutationSymmetryInfo(pydantic.BaseModel):
    symmetry: Dict[str, young.Partition]
    n_terms: pydantic.PositiveInt
    n_operators: pydantic.PositiveInt
    matrix: utils.IntegerMatrix

    def __tex__(self):
        sym = (
            "$$"
            + r"\,,\quad".join(
                rf"\lambda_{{${utils.esc(f)}}}\sim{p}" for f, p in self.symmetry.items()
            )
            + "$$"
        )
        ext = (
            rf"$$\mathcal{{N}}_{{\text{{terms}}}}={self.n_terms}$$"
            r"\quad\&\quad"
            rf"$$\mathcal{{N}}_{{\text{{operators}}}}={self.n_operators}$$"
        )
        mat = "\\\\\n".join(
            " & ".join(str(entry) for entry in row) for row in self.matrix
        )
        mat = f"\\begin{{pmatrix}}\n{mat}\n\\end{{pmatrix}}"
        return (
            f"{sym}\n\\hfill{{\\small{ext}}}\n"
            f"\\begin{{gather}}\n\\mathcal{{K}}=\n{mat}\n\\end{{gather}}"
        )


class OperatorInfoPermutation(OperatorInfo):
    permutation_symmetries: List[PermutationSymmetryInfo]

    def __tex__(self):
        tex_fields = {
            utils.esc(f): tex(field) for f, field in self.model.fields.items()
        }
        content = (
            rf"\item {string.Template(tex(sym)).substitute(tex_fields)}"
            for sym in self.permutation_symmetries
        )
        items = "\n".join((r"\begin{itemize}", *content, r"\end{itemize}"))
        all_items = super().__tex__().split("\n")
        return "\n".join(
            (
                *all_items[:-1],
                r"\item[permutation symmetries]\hfill",
                items,
                all_items[-1],
            ),
        )

    @pydantic.validator("invariants", pre=True)
    def parse_invariants(cls, v, values):
        invariants = defaultdict(list)
        if "model" in values:
            model = values["model"]
            lorentz_group = model.symmetries.lorentz_group
            for group_name, contractions in v.items():
                for contraction in contractions.values():
                    if group_name == lorentz_group.name:
                        tab, sign = tensors.LorentzTableau.parse_tableau(contraction)
                        if not tab.is_ssyt():
                            errmsg = (
                                "Invalid tableau encountered in operator file."
                                "\n"
                                "Please make sure the file contains only Lorentz"
                                " structures that correspond to SSYTx."
                            )
                            raise ValueError(errmsg)
                        if sign != 1:
                            errmsg = (
                                "Invalid sign encountered in operator file."
                                "\n"
                                "Please make sure the file contains only Lorentz"
                                " structures that correspond to SSYTx."
                            )
                            raise ValueError(errmsg)
                    else:
                        tab, sign = tensors.SUNTableau.parse_tableau(contraction)
                        if sign != 1:
                            errmsg = (
                                "Invalid sign encountered in operator file."
                                "\n"
                                "Please make sure the file contains only SU(N)"
                                " structures with positive signs."
                            )
                            raise ValueError(errmsg)
                    invariants[group_name].append((tab, sign))
        return invariants

    @pydantic.validator("permutation_symmetries", pre=True)
    def parse_permutation_symmetries(cls, v, values):
        if "model" in values:
            model = values["model"]
            lorentz_group = model.symmetries.lorentz_group
            sun_groups = model.symmetries.sun_groups
            if v[0]["vector"].split(" * ") != [lorentz_group.name, *sun_groups.keys()]:
                errmsg = (
                    "Uncompatible tensor product encountered in operator file."
                    "\n"
                    "Please make sure the file contains the same order of symmetries"
                    " as the IR model."
                )
                raise ValueError(errmsg)
        return v[1:]

    def yaml_dict(self):
        operator_data = super().yaml_dict()
        permutations_data = [utils.BlockMap(vector=" * ".join(self.invariants.keys()))]
        for permutation_symmetry in self.permutation_symmetries:
            permutation_data = {
                "symmetry": utils.FlowMap(permutation_symmetry.symmetry),
            }
            permutation_data.update(permutation_symmetry.dict(exclude={"symmetry"}))
            permutations_data.append(permutation_data)
        operator_data["permutation_symmetries"] = permutations_data
        return operator_data

    def expanded(self) -> "OperatorInfoExpanded":
        return OperatorInfoExpanded.from_op_info_permutation(self)


@dataclass
class TermInfo:
    op_type: classes.Type
    symmetry: Dict[str, young.Partition]
    n_operators: pydantic.PositiveInt
    term: tensors.CombinedTensor
    operators: List[Tuple[int, ...]]

    def __str__(self):
        return self.term.render_term(self.op_type)


class ExpandedInfo(pydantic.BaseModel):
    symmetry: Dict[str, young.Partition]
    n_terms: pydantic.PositiveInt
    n_operators: pydantic.PositiveInt
    terms: List[tensors.CombinedTensor]
    operators: List[Tuple[int, ...]]


class OperatorInfoExpanded(OperatorInfo):
    expanded: List[ExpandedInfo]

    def __iter__(self):
        return iter(
            TermInfo(
                op_type=self.op_type,
                symmetry=ex_info.symmetry,
                n_operators=len(ex_info.operators),
                term=term,
                operators=ex_info.operators,
            )
            for ex_info in self.expanded
            for term in ex_info.terms
        )

    def yaml_dict(self):
        operator_data = super().yaml_dict()
        expanded_data = []
        for permutation_symmetry in self.expanded:
            permutation_data = {
                "symmetry": utils.FlowMap(permutation_symmetry.symmetry),
            }
            permutation_data.update(
                permutation_symmetry.dict(exclude={"symmetry", "terms", "operators"}),
            )
            permutation_data["terms"] = [
                term.render_term(self.op_type) for term in permutation_symmetry.terms
            ]
            permutation_data["operators"] = permutation_symmetry.operators
            expanded_data.append(permutation_data)
        operator_data["expanded"] = expanded_data
        return operator_data

    @classmethod
    def from_op_info_permutation(cls, op_info) -> "OperatorInfoExpanded":
        if isinstance(op_info, OperatorInfoPermutation):
            lengths = {
                group: len(tensors) for group, tensors in op_info.invariants.items()
            }
            expanded = []
            for permutation_symmetry in op_info.permutation_symmetries:
                gens = []
                for f, symmetry in permutation_symmetry.symmetry.items():
                    field = op_info.model.fields[f]
                    gens.append(
                        [
                            tab.flatten()
                            for tab in young.Partition(symmetry).semistandard_tableaux(
                                field.generations,
                            )
                        ],
                    )
                operators = [
                    tuple(index for indices in operator for index in indices)
                    for operator in itertools.product(*gens)
                ]
                terms = []
                for row in permutation_symmetry.matrix:
                    term = tensors.CombinedTensor()
                    for n, coeff in enumerate(row):
                        if coeff:
                            tabs = {}
                            term_coeff = 1
                            den = 1
                            for group in reversed(op_info.invariants):
                                tab, tab_coeff = op_info.invariants[group][
                                    (n // den) % lengths[group]
                                ]
                                tabs[group] = tab
                                term_coeff *= tab_coeff
                                den *= lengths[group]
                            term[tensors.CombinedTableaux(reversed(tabs.items()))] = (
                                coeff * term_coeff
                            )
                    terms.append(term)
                expanded.append(
                    ExpandedInfo.construct(
                        symmetry=permutation_symmetry.symmetry,
                        n_terms=permutation_symmetry.n_terms,
                        n_operators=permutation_symmetry.n_operators,
                        terms=terms,
                        operators=operators,
                    ),
                )
            return cls.construct(
                model=op_info.model,
                version=op_info.version,
                type=op_info.type,
                generations=op_info.generations,
                n_terms=op_info.n_terms,
                n_operators=op_info.n_operators,
                invariants=op_info.invariants,
                expanded=expanded,
            )
        raise NotImplementedError


class Basis(dict):
    model: ir_model.Model

    def __init__(self, model: ir_model.Model):
        super().__init__()
        self.model = model

    def __getitem__(self, key):
        if isinstance(key, (int, slice)):
            return list(self.values())[key]
        if isinstance(key, dict):
            op_type = classes.Type.from_counter(key, self.model)
            return self[op_type]
        if isinstance(key, classes.Type):
            return self[repr(key)]
        return super().__getitem__(key)



import itertools
from collections import Counter
from math import ceil
from math import floor
from pathlib import Path
from typing import Counter as Count
from typing import Dict
from typing import List
from typing import Tuple

import autoeft.base.groups as symmetry_groups
import autoeft.base.model as ir_model
import autoeft.base.representations as reprs
from autoeft.combinatorics import young
from autoeft.tex import tex


class Family(tuple):
    hel2 = (-2, -1, 0, 1, 2)
    _repr = ("FL", "psiL", "phi", "psiR", "FR", "D")
    _str = ("FL", "psiL", "phi", "psiR", "FR", "D")
    _tex = (
        r"F_{\mathrm{L}}",
        r"\psi_{\mathrm{L}}",
        r"\phi",
        r"\psi_{\mathrm{R}}",
        r"F_{\mathrm{R}}",
        r"D",
    )

    def __new__(cls, n_hel, nD):
        return super().__new__(cls, (*n_hel, nD))

    def __repr__(self):
        return "_".join(f"{m}{f}" for f, m in zip(self._repr, self) if m)

    def __str__(self):
        return " ".join(f"{f}({m})" for f, m in zip(self._str, self) if m)

    def __tex__(self):
        return " ".join(
            f"{{{f}}}^{{{m}}}" if m > 1 else f for f, m in zip(self._tex, self) if m
        )

    @property
    def n_hel(self) -> Tuple[int, ...]:
        return self[:-1]

    @property
    def nD(self) -> int:
        return self[-1]

    @property
    def N(self) -> int:
        return sum(self.n_hel)

    @property
    def nl(self) -> int:
        left = floor(len(self.n_hel) / 2)
        nl2 = sum(
            (abs(h2) * nh for h2, nh in zip(self.hel2[:left], self.n_hel[:left])),
            start=self.nD,
        )
        assert nl2 % 2 == 0
        return nl2 // 2

    @property
    def nr(self) -> int:
        right = ceil(len(self.n_hel) / 2)
        nr2 = sum(
            (abs(h2) * nh for h2, nh in zip(self.hel2[right:], self.n_hel[right:])),
            start=self.nD,
        )
        assert nr2 % 2 == 0
        return nr2 // 2

    @property
    def d(self) -> int:
        return self.N + self.nl + self.nr

    def tuple_repr(self) -> str:
        return "(" + ", ".join(str(nh) for nh in self.n_hel) + "; " + str(self.nD) + ")"

    def path(self) -> Path:
        return Path(ascii(self.N)) / f"{self!a}"

    def is_real(self) -> bool:
        """Return true if is equal to its own conjugate."""
        return list(self.n_hel) == list(reversed(self.n_hel))

    def is_complex(self) -> bool:
        """Return true if is unequal to its own conjugate."""
        return not self.is_real()

    def is_normal(self) -> bool:
        return list(self.n_hel) >= list(reversed(self.n_hel))

    def conjugated(self) -> "Family":
        return type(self)(reversed(self.n_hel), self.nD)

    def field_helicities2(self) -> List[int]:
        """Return a list of the doubled field helicities."""
        return list(
            itertools.chain.from_iterable(
                itertools.repeat(h2, nh) for h2, nh in zip(self.hel2, self.n_hel)
            ),
        )

    def primary_partition(self) -> young.Partition:
        """Return the primary partition.

        The primary partition is defined by the Young diagram with
        nr columns of length N - 2 and nl columns of length 2.
        """
        if self.N == 3 and self.nl == 0:
            return young.Partition([self.nr])
        return young.Partition([self.nl + self.nr] * 2 + [self.nr] * (self.N - 4))

    def primary_content(self) -> Tuple[int, ...]:
        return tuple(self.nr - h2 for h2 in self.field_helicities2())


class FamilyGR(Family):
    hel2 = (-4, -2, -1, 0, 1, 2, 4)
    _repr = ("CL", "FL", "psiL", "phi", "psiR", "FR", "CR", "D")
    _str = ("CL", "FL", "psiL", "phi", "psiR", "FR", "CR", "D")
    _tex = (
        r"C_{\mathrm{L}}",
        r"F_{\mathrm{L}}",
        r"\psi_{\mathrm{L}}",
        r"\phi",
        r"\psi_{\mathrm{R}}",
        r"F_{\mathrm{R}}",
        r"C_{\mathrm{R}}",
        r"D",
    )

    @property
    def nGR(self) -> int:
        return self.n_hel[0] + self.n_hel[-1]

    @property
    def d(self) -> int:
        return self.N + self.nl + self.nr - self.nGR


class Type(tuple):
    family: Family

    def __new__(
        cls,
        field_counts: Count[ir_model.Field],
        _,
    ):
        if field_counts is None:
            return super().__new__(cls)
        return super().__new__(
            cls,
            sorted(
                ((f, m) for f, m in field_counts.items() if m),
                key=lambda x: (
                    x[0].helicity,
                    x[0].name,
                ),
            ),
        )

    def __init__(
        self,
        _,
        family: Family,
    ) -> None:
        self.family = family

    def __repr__(self):
        return "_".join(f"{m}{f}" for f, m in self.with_deriv if m)

    def __str__(self):
        return " ".join(f"{f}({m})" for f, m in self.with_deriv if m)

    def __tex__(self):
        return " ".join(
            f"{{{tex(f)}}}^{{{m}}}" if m > 1 else f"{tex(f)}"
            for f, m in self.with_deriv
            if m
        )

    @property
    def nD(self) -> int:
        return self.family.nD

    @property
    def d(self) -> int:
        return self.family.d

    @property
    def with_deriv(self):
        return (*self, ("D", self.nD))

    def path(self, extension: str = "yml") -> Path:
        return Path(f"{self!a}.{extension}")

    def full_path(self, extension: str = "yml") -> Path:
        return self.family.path() / self.path(extension)

    def is_real(self) -> bool:
        """Return true if is equal to its own conjugate."""
        return self.family.is_real() and self == self.conjugated()

    def is_complex(self) -> bool:
        """Return true if is unequal to its own conjugate."""
        return not self.is_real()

    def is_normal(self) -> bool:
        if self.family.is_normal():
            if self.family.is_real():
                normal_form = tuple((f.name, m) for f, m in self)
                cnormal_form = tuple((f.name, m) for f, m in self.conjugated())
                return self.family.is_normal() and normal_form <= cnormal_form
            return True
        return False

    def conjugated(self) -> "Type":
        return type(self)(
            {f.hermitian_conjugated(): m for f, m in self},
            self.family.conjugated(),
        )

    def sun_partition(self, group: symmetry_groups.SUNGroup) -> young.Partition:
        total = sum(sum(f.representations[group.name]) * m for f, m in self)
        assert total % group.N == 0
        return young.Partition([total // group.N] * group.N)

    def sun_content(self, group: symmetry_groups.SUNGroup) -> List[reprs.SUNRepr]:
        return list(
            itertools.chain.from_iterable(
                itertools.repeat(f.representations[group.name], m) for f, m in self
            ),
        )

    def field_numbers(self) -> Dict[str, List[int]]:
        field_number = 1
        return {
            f.name: list(range(field_number, field_number := field_number + m))
            for f, m in self
        }

    def to_counter(self) -> Counter:
        return +Counter({str(f): m for f, m in self.with_deriv})

    @classmethod
    def from_counter(cls, field_content: Count[str], model: ir_model.Model) -> "Type":
        field_content = field_content.copy()
        nD = field_content.pop("D", 0)
        hel2_counts = Counter()
        field_counts = Counter()
        gr_flag = False
        for field_name, multiplicity in field_content.items():
            field = model.fields[field_name]
            hel2 = int(2 * field.helicity)
            hel2_counts[hel2] += multiplicity
            field_counts[field] += multiplicity
            if abs(hel2) == 4:
                gr_flag = True
        F = FamilyGR if gr_flag else Family
        assert set(hel2_counts) <= set(F.hel2)
        n_hel = (hel2_counts[hel2] for hel2 in F.hel2)
        family = F(n_hel, nD)
        return cls(field_counts, family)



import importlib.metadata
import importlib.resources
import os
import sys
from pathlib import Path

import packaging.version
import rich.console
import rich.text
import rich.theme
import semver

from autoeft import exceptions

program = "AutoEFT"
__program__ = program.lower()

# load metadata
try:
    dist = importlib.metadata.distribution(__program__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "<version>"
    __authors__ = "<authors>"
    __license__ = "<license>"
    __summary__ = "<summary>"
    __webpage__ = "<url>"
else:
    __program__ = dist.metadata["Name"]
    __version__ = dist.metadata["Version"]
    __authors__ = dist.metadata["Author"]
    __license__ = dist.metadata["License"].split("\n")[0]
    __summary__ = dist.metadata["Summary"]
    __webpage__ = dist.metadata["Project-URL"]

pkg_version = packaging.version.Version(__version__)
version = semver.Version(
    *pkg_version.release,
    prerelease=(
        None if pkg_version.pre is None else "".join([str(i) for i in pkg_version.pre])
    ),
    build=None if pkg_version.dev is None else f"dev{pkg_version.dev}",
)

# load resources
try:
    if sys.version_info >= (3, 9):
        __copyright__ = (
            importlib.resources.files(__program__).joinpath("COPYRIGHT").read_text()
        )
        __logo__ = importlib.resources.files(__program__).joinpath("LOGO").read_text()
    else:
        __copyright__ = importlib.resources.read_text(__program__, "COPYRIGHT")
        __logo__ = importlib.resources.read_text(__program__, "LOGO")
except FileNotFoundError:
    __copyright__ = "<copyright>\n"
    __logo__ = "<LOGO>\n"

try:
    if sys.version_info >= (3, 9):
        with importlib.resources.as_file(
            importlib.resources.files(__program__).joinpath("theme.ini"),
        ) as theme_path:
            theme = rich.theme.Theme().read(str(theme_path))
    else:
        with importlib.resources.path(__program__, "theme.ini") as theme_path:
            theme = rich.theme.Theme().read(str(theme_path))
except FileNotFoundError:
    theme = rich.theme.Theme()


console = rich.console.Console(theme=theme)
log_file = Path.cwd() / "autoeft.log"
log_console = rich.console.Console(file=log_file.open("w"), width=80, log_path=False)


def print(*args, dest: str = "both"):  # noqa: A001
    if dest == "both":
        console.print(*args)
        log_console.log(*args)
    elif dest == "terminal":
        console.print(*args)
    elif dest == "log":
        log_console.log(*args)
    else:
        raise ValueError(dest)


logo = rich.text.Text(__logo__)
logo.highlight_regex("[Auto]", style="bold rgb(142,186,229)")
logo.highlight_regex("[EFT]", style="bold rgb(0,84,159)")

disclaimer = rich.text.Text()
disclaimer.append(program, style="bold")
disclaimer.append(" ")
disclaimer.append(str(version), style="italic")
disclaimer.append("\n")
disclaimer.append(__copyright__)
disclaimer.append("This application is licensed under the ")
disclaimer.append(__license__, style="italic")
disclaimer.append(".")

conjugation_symbol = os.getenv("AUTOEFT_CS", "+")
epsilon_dot_symbol = os.getenv("AUTOEFT_DS", "~")

try:
    from sage.version import version as __sage_version__

    sage_version_min = semver.Version(9, 3)
    sage_version = semver.Version.parse(__sage_version__, optional_minor_and_patch=True)
    if sage_version < sage_version_min:
        errmsg = (
            f"The version of SageMath installed is not compatible with {program}."
            "\n"
            f"Please install SageMath version {sage_version_min} or higher."
        )
        raise exceptions.RequirementVersionError(errmsg)
except ModuleNotFoundError as e:
    errmsg = (
        "Unable to import SageMath."
        "\n"
        "Please make sure the module is installed and importable."
    )
    raise exceptions.MissingRequirementError(errmsg) from e



import itertools
from collections import defaultdict
from typing import Counter as Count
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple

import autoeft.base.classifications as classes
import autoeft.base.model as ir_model
import autoeft.combinatorics.symmetric_group as combinat
import autoeft.construction.symmetrizers as sym_reprs
from autoeft.combinatorics import young

GroupName = FieldName = str
DirectProduct = Dict[GroupName, Dict[FieldName, young.Partition]]
ProductDecomposition = Dict[FieldName, Count[young.Partition]]
ClebschGordanDecomposition = Dict[
    FieldName,
    Dict[young.Partition, List[Count[Tuple[int, ...]]]],
]


def construct_factorized_symmetries(
    direct_product: DirectProduct,
    product_decomposition: ProductDecomposition,
    clebsch_gordan_decomposition: ClebschGordanDecomposition,
) -> Iterator[
    Tuple[
        Dict[FieldName, young.Partition],
        Count[Tuple[sym_reprs.GroupSymmetrizer, ...]],
    ]
]:
    total_permutation_symmetries = (
        dict(zip(product_decomposition.keys(), symmetry))
        for symmetry in itertools.product(*product_decomposition.values())
    )
    for permutation_symmetry in total_permutation_symmetries:
        multiplicities = itertools.product(
            *(
                range(int(product_decomposition[field_name][permutation]))
                for field_name, permutation in permutation_symmetry.items()
            ),
        )
        for multiplicity in multiplicities:
            cgcs = (
                clebsch_gordan_decomposition[field_name][symmetry][j]
                for j, (field_name, symmetry) in zip(
                    multiplicity,
                    permutation_symmetry.items(),
                )
            )
            combination = {
                tuple(
                    sym_reprs.GroupSymmetrizer(
                        group_name,
                        tuple(
                            (field_name, group_permutation, entry)
                            for (field_name, group_permutation), entry in zip(
                                group_permutations.items(),
                                vector,
                            )
                        ),
                    )
                    for (group_name, group_permutations), vector in zip(
                        direct_product.items(),
                        vectors,
                    )
                ): cgc
                for vectors, cgc in combinat.join_CGCs(cgcs).items()
                if cgc
            }
            yield permutation_symmetry, combination


def construct_permutation_symmetries(
    op_type: classes.Type,
    allowed_group_permutations: Dict[GroupName, List[Dict[FieldName, young.Partition]]],
) -> Iterator[Tuple[DirectProduct, ProductDecomposition, ClebschGordanDecomposition]]:
    combined_permutations = (
        dict(zip(allowed_group_permutations.keys(), combined_permutation))
        for combined_permutation in itertools.product(
            *allowed_group_permutations.values(),
        )
    )
    for permutation in combined_permutations:
        field_permutations = defaultdict(dict)
        for group_name, allowed_permutation in permutation.items():
            for field_name, group_permutation in allowed_permutation.items():
                field_permutations[field_name][group_name] = group_permutation
        permutation_symmetry = {}
        clebsch_gordan_coefficients = defaultdict(dict)
        for f, m in op_type:
            partitions = list(field_permutations[f.name].values())
            if bool(f.anticommute) ^ bool(op_type.family.nr % 2):
                partitions.append(young.Partition([1] * m))
            targets = permutation_symmetry[f.name] = {
                symmetry: mult
                for symmetry, mult in combinat.multi_direct_product_decomposition(
                    partitions,
                    m,
                ).items()
                if len(symmetry) <= f.generations
            }
            for target, multiplicity in targets.items():
                matrix, basis = combinat.construct_CGCs(
                    m,
                    partitions,
                    target,
                    multiplicity,
                )
                if bool(f.anticommute) ^ bool(op_type.family.nr % 2):
                    basis = [vector[:-1] for vector in basis]
                clebsch_gordan_coefficients[f.name][target] = [
                    dict(zip(basis, vector)) for vector in matrix
                ]
        yield permutation, permutation_symmetry, clebsch_gordan_coefficients


def construct_group_permutations(
    op_type: classes.Type,
    symmetries: ir_model.Symmetries,
) -> Dict[GroupName, List[Dict[FieldName, young.Partition]]]:
    family = op_type.family
    lorentz_group = symmetries.lorentz_group
    sun_groups = symmetries.sun_groups
    targets = {}
    if target := family.primary_partition():
        targets[lorentz_group.name] = target
    for group_name, group in sun_groups.items():
        if target := op_type.sun_partition(group):
            targets[group_name] = target
    representations = defaultdict(dict)
    permutations = defaultdict(dict)
    for f, m in op_type:
        lorentz_repr = young.Partition([int(family.nr - 2 * f.helicity)])
        representations[lorentz_group.name][f.name] = lorentz_repr
        permutations[lorentz_group.name][f.name] = young.Partition.partitions(m)
        for group_name in sun_groups:
            sun_repr = f.representations[group_name].partition
            representations[group_name][f.name] = sun_repr
            permutations[group_name][f.name] = young.Partition.partitions(m)
    allowed_permutations = defaultdict(list)
    for group_name, target in targets.items():
        all_permutations = (
            dict(zip(permutations[group_name].keys(), permutation))
            for permutation in itertools.product(*permutations[group_name].values())
        )
        for perm in all_permutations:
            for partitions in itertools.product(
                *(
                    combinat.plethysm(field_repr, perm[field_name]).keys()
                    for field_name, field_repr in representations[group_name].items()
                ),
            ):
                if target in combinat.multi_lr_rule(partitions, None):
                    allowed_permutations[group_name].append(perm)
                    break
    return allowed_permutations



import concurrent.futures
import itertools
import os
import re
from collections import Counter
from dataclasses import dataclass
from itertools import zip_longest
from typing import Callable
from typing import Counter as Count
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

import sage.all
import sage.combinat.permutation
import sage.combinat.symmetric_group_algebra
import sage.matrix.constructor
import sage.modules.free_module_element
from sage.rings.rational_field import QQ

import autoeft
import autoeft.base.classifications as classes
import autoeft.base.groups as symmetry_groups
import autoeft.base.representations as reprs
from autoeft import utils
from autoeft.combinatorics import young
from autoeft.form import contract
from autoeft.form import form

eps_undotted_pattern = re.compile(r"^eps\((\d+)_\d+,(\d+)_\d+\)$")
eps_dotted_pattern = re.compile(
    rf"^eps\((\d+)_\d+{re.escape(autoeft.epsilon_dot_symbol)}"
    rf",(\d+)_\d+{re.escape(autoeft.epsilon_dot_symbol)}\)$",
)
eps_pattern = re.compile(r"^eps\((.*)\)$")


_Permutations = sage.combinat.permutation.Permutations()


# Lorentz structures


class LorentzTensor(utils.Vector):
    def symmetrize(self, symmetrizer: Counter) -> "LorentzTensor":
        res = sum(
            (coeff * tab.symmetrized(symmetrizer) for tab, coeff in self.items()),
            start=LorentzTensor(),
        )
        self.clear()
        self.update(res)
        return self

    def reduction(
        self,
        basis: List["LorentzTableau"],
        N: int,
        n_split: int,
    ) -> "LorentzTensor":
        coeffs = Counter()
        for tab, factor in self.items():
            lr_tab, lr_sign = tab.to_lr_tableau()
            coeffs.update(ibp_reduction(lr_tab, lr_sign * factor))
        comb = LorentzTensor()
        for tab, coeff in coeffs.items():
            ten, lr_sign = tab.to_lorentz_tableau()
            comb[ten] = coeff * lr_sign
        # assert that every tableau is reduced to a combination of SSYTx only
        assert set(comb) <= set(basis)
        return comb


class LorentzTableau(young.Tableau):
    N: int
    n_split: int

    def __new__(cls, _, __, tableau: Optional[Iterable[Iterable]] = None):
        if tableau:
            return super().__new__(
                cls,
                ((int(index) for index in row) for row in tableau),
            )
        return super().__new__(cls, tableau)

    def __init__(self, N: int, n_split: int, _=None):
        self.N = N
        self.n_split = n_split

    @property
    def sign(self):
        _, lr_sign = self.to_lr_tableau()
        return lr_sign

    @property
    def undotted_indices(self) -> Counter:
        lr_tab, _ = self.to_lr_tableau()
        return lr_tab.undotted_indices

    @property
    def dotted_indices(self) -> Counter:
        lr_tab, _ = self.to_lr_tableau()
        return lr_tab.dotted_indices

    def permuted(self, perm: Tuple[int, ...]) -> "LorentzTableau":
        perm_map = dict(zip(sorted(perm), perm))
        return LorentzTableau(
            self.N,
            self.n_split,
            ((perm_map.get(index, index) for index in row) for row in self),
        )

    def symmetrized(self, symmetrizer: Counter) -> LorentzTensor:
        return sum(
            (
                LorentzTensor({self.permuted(perm): coeff})
                for perm, coeff in symmetrizer.items()
            ),
            start=LorentzTensor(),
        )

    def to_lr_tableau(self) -> Tuple["LRTableau", int]:
        l_tab = self[:, self.n_split :]
        r_tab_c = self[:, : self.n_split]
        c_columns, lc_sign = [], 1
        for column in reversed(r_tab_c.transposed()):
            c_columns.append(
                c_column := [i for i in range(1, self.N + 1) if i not in column],
            )
            lc_sign *= _Permutations(list(column) + c_column).sign()
        r_tab = young.Tableau(row for row in zip(*c_columns))
        return LRTableau(self.N, l_tab, r_tab), lc_sign

    def render_eps(self, prefix: str = "") -> str:
        lr_tab, _ = self.to_lr_tableau()
        return lr_tab.render_eps(prefix=prefix)

    def render_tableau(self, op_type: classes.Type, extra_sign: int = 1) -> str:
        sign = "+" if self.sign * extra_sign == 1 else "-"
        tensor = self.render_eps() or "1"
        fields = []
        field_num = 0
        for f, m in op_type:
            for i in range(1, m + 1):
                n_undotted = self.undotted_indices[field_num + i]
                n_dotted = self.dotted_indices[field_num + i]
                nD = min(n_undotted, n_dotted)
                if nD == 1:
                    field = f"(D {f.name})"
                elif nD > 1:
                    field = f"(D^{nD} {f.name})"
                else:
                    field = f.name
                if n_undotted or n_dotted:
                    undotted = [f"{field_num+i}_{j}" for j in range(1, n_undotted + 1)]
                    dotted = [
                        f"{field_num+i}_{j}{autoeft.epsilon_dot_symbol}"
                        for j in range(1, n_dotted + 1)
                    ]
                    field_indices = ",".join(undotted + dotted)
                    field += f"({field_indices})"
                fields.append(field)
            field_num += m
        fields = "*".join(fields)
        return sign + tensor + " * " + fields

    def render_eps_tex(self, lorentz_group: symmetry_groups.LorentzGroup) -> str:
        lr_tab, _ = self.to_lr_tableau()
        return lr_tab.render_eps_tex(lorentz_group)

    def render_tableau_tex(
        self,
        op_type: classes.Type,
        lorentz_group: symmetry_groups.LorentzGroup,
        extra_sign: int = 1,
    ) -> str:
        sign = "+" if self.sign * extra_sign == 1 else "-"
        tensor = self.render_eps_tex(lorentz_group) or "1"
        fields = []
        field_num = 0
        for f, m in op_type:
            for i in range(1, m + 1):
                n_undotted = self.undotted_indices[field_num + i]
                n_dotted = self.dotted_indices[field_num + i]
                nD = min(n_undotted, n_dotted)
                if nD == 1:
                    field = rf"(D\,{f.tex})"
                elif nD > 1:
                    field = rf"(D^{{{nD}}}\,{f.tex})"
                else:
                    field = f.tex
                if n_undotted or n_dotted:
                    undotted = ",".join(
                        lorentz_group.get_l_idx(field_num + i, j)
                        for j in range(1, n_undotted + 1)
                    )
                    dotted = ",".join(
                        lorentz_group.get_r_idx(field_num + i, j)
                        for j in range(1, n_dotted + 1)
                    )
                    field = f"{{{field}}}_{{{undotted}}}^{{{dotted}}}"
                fields.append(field)
            field_num += m
        fields = r"\,".join(fields)
        return rf"{sign} \, {tensor} \; {fields}"

    @classmethod
    def parse_tableau(cls, invariant: str) -> Tuple["LorentzTableau", int]:
        split_invariant = invariant.split(" * ")
        signed_tensor = split_invariant[0]
        fields = split_invariant[1]
        if signed_tensor == "+1":
            return LorentzTableau(0, 0), 1
        if signed_tensor == "-1":
            return LorentzTableau(0, 0), -1
        sign = signed_tensor[0]
        tensor = signed_tensor[1:]
        epsilons = tensor.split("*")
        left_tab = []
        right_tab = []
        for eps in epsilons:
            if match := eps_undotted_pattern.match(eps):
                left_tab.append(map(int, match.groups()))
            elif match := eps_dotted_pattern.match(eps):
                right_tab.append(map(int, match.groups()))
            else:
                errmsg = (
                    "Invalid Lorentz tensor."
                    "\n"
                    "Tensors must be a product of 'eps(i_j,k_l)'."
                    "\n"
                    f"{eps}"
                )
                raise ValueError(errmsg)
        N = len(fields.split("*"))
        lr_tab = LRTableau(
            N,
            young.Tableau(zip(*left_tab)),
            young.Tableau(zip(*right_tab)),
        )
        tab, lc_sign = lr_tab.to_lorentz_tableau()
        sign_map = {"+": 1, "-": -1}
        return tab, sign_map[sign] * lc_sign


@dataclass(frozen=True)
class LRTableau:
    N: int
    left_tableau: young.Tableau
    right_tableau: young.Tableau

    @property
    def nl(self):
        if self.left_tableau:
            return len(self.left_tableau[0])
        return 0

    @property
    def nr(self):
        if self.right_tableau:
            return len(self.right_tableau[0])
        return 0

    @property
    def undotted_indices(self) -> Counter:
        return Counter(index for row in self.left_tableau for index in row)

    @property
    def dotted_indices(self) -> Counter:
        return Counter(index for row in self.right_tableau for index in row)

    def to_lorentz_tableau(self) -> Tuple[LorentzTableau, int]:
        columns, lc_sign = [], 1
        for c_column in reversed(self.right_tableau.transposed()):
            columns.append(
                column := [i for i in range(1, self.N + 1) if i not in c_column],
            )
            lc_sign *= _Permutations(column + list(c_column)).sign()
        ltab = LorentzTableau(
            self.N,
            self.nr,
            (
                r_row + l_row
                for r_row, l_row in zip_longest(
                    zip(*columns),
                    self.left_tableau,
                    fillvalue=(),
                )
            ),
        )
        return ltab, lc_sign

    def normalized(self) -> Tuple["LRTableau", int]:
        def _normalized(tableau: young.Tableau) -> Tuple[young.Tableau, int]:
            normalized_tableau = tableau.to_lists()
            normalization_sign = 1
            for i, column in enumerate(tableau.transposed()):
                if column[0] == column[1]:
                    normalization_sign = 0
                elif column[0] > column[1]:
                    normalized_tableau[0][i], normalized_tableau[1][i] = (
                        tableau[1, i],
                        tableau[0, i],
                    )
                    normalization_sign *= -1
            return (
                young.Tableau(zip(*sorted(zip(*normalized_tableau)))),
                normalization_sign,
            )

        left_tableau, l_sign = _normalized(self.left_tableau)
        right_tableau, r_sign = _normalized(self.right_tableau)
        return LRTableau(self.N, left_tableau, right_tableau), l_sign * r_sign

    def is_sorted(self) -> bool:
        is_sorted_left = is_sorted_right = True
        if self.left_tableau:
            is_sorted_left = list(self.left_tableau[1]) == sorted(self.left_tableau[1])
        if self.right_tableau:
            is_sorted_right = list(self.right_tableau[1]) == sorted(
                self.right_tableau[1],
            )
        return is_sorted_left and is_sorted_right

    def schouten(self) -> Count["LRTableau"]:
        def _schouten(tableau: young.Tableau, c: int = 0) -> Count[young.Tableau]:
            cc = c + 1
            ccc = cc + 1
            if not tableau or cc >= len(tableau[0]):
                return Counter({tableau: 1})
            if tableau[0, c] < tableau[0, cc] and tableau[1, c] > tableau[1, cc]:
                result = Counter()
                result.update(
                    _schouten(
                        young.Tableau(
                            (
                                tableau[0],
                                (
                                    *tableau[1, :c],
                                    tableau[1, cc],
                                    tableau[1, c],
                                    *tableau[1, ccc:],
                                ),
                            ),
                        ),
                    ),
                )
                result.subtract(
                    _schouten(
                        young.Tableau(
                            (
                                (*tableau[0, :cc], tableau[1, cc], *tableau[0, ccc:]),
                                (
                                    *tableau[1, :c],
                                    tableau[0, cc],
                                    tableau[1, c],
                                    *tableau[1, ccc:],
                                ),
                            ),
                        ),
                    ),
                )
                return result
            return _schouten(tableau, cc)

        result = Counter()
        if self.is_sorted():
            return result
        left_tableaux = _schouten(self.left_tableau)
        right_tableaux = _schouten(self.right_tableau)
        for (l_tab, l_coeff), (r_tab, r_coeff) in itertools.product(
            left_tableaux.items(),
            right_tableaux.items(),
        ):
            result[LRTableau(self.N, l_tab, r_tab)] += l_coeff * r_coeff

        return result

    def ibp1(
        self,
        coeff: Optional[utils.Fraction] = None,
    ) -> Count["LRTableau"]:
        if coeff is None:
            coeff = utils.Fraction(1)
        result = Counter()
        l_tab, r_tab = self.left_tableau, self.right_tableau
        for i in range(2, self.N + 1):
            combination = list(
                itertools.product(l_tab.replaced(1, i), r_tab.replaced(1, i)),
            )
            for sl_tab, sr_tab in combination:
                result[LRTableau(self.N, sl_tab, sr_tab)] -= coeff / len(combination)
        return result

    def ibp2a(
        self,
        coeff: Optional[utils.Fraction] = None,
    ) -> Count["LRTableau"]:
        if coeff is None:
            coeff = utils.Fraction(1)
        result = Counter()
        l_tab, r_tab = self.left_tableau, self.right_tableau
        l_tab_T = list(zip(*l_tab))
        try:
            idx = l_tab_T.index((1, 2))
        except ValueError:
            pass
        else:
            for i in range(3, self.N + 1):
                l_tab_T[idx] = (1, i)
                sl_tab = young.Tableau(zip(*l_tab_T))
                sr_tabs = list(r_tab.replaced(2, i))
                for sr_tab in sr_tabs:
                    result[LRTableau(self.N, sl_tab, sr_tab)] -= coeff / len(sr_tabs)
        return result

    def ibp2b(
        self,
        coeff: Optional[utils.Fraction] = None,
    ) -> Count["LRTableau"]:
        if coeff is None:
            coeff = utils.Fraction(1)
        result = Counter()
        l_tab, r_tab = self.left_tableau, self.right_tableau
        r_tab_T = list(zip(*r_tab))
        try:
            idx = r_tab_T.index((1, 2))
        except ValueError:
            pass
        else:
            for i in range(3, self.N + 1):
                r_tab_T[idx] = (1, i)
                sr_tab = young.Tableau(zip(*r_tab_T))
                sl_tabs = list(l_tab.replaced(2, i))
                for sl_tab in sl_tabs:
                    result[LRTableau(self.N, sl_tab, sr_tab)] -= coeff / len(sl_tabs)
        return result

    def ibp3a(
        self,
        coeff: Optional[utils.Fraction] = None,
    ) -> Count["LRTableau"]:
        if coeff is None:
            coeff = utils.Fraction(1)
        result = Counter()
        l_tab, r_tab = self.left_tableau, self.right_tableau
        l_tab_T, r_tab_T = list(zip(*l_tab)), list(zip(*r_tab))
        try:
            l_idx = l_tab_T.index((1, 3))
            r_idx = r_tab_T.index((2, 3))
        except ValueError:
            pass
        else:
            for i in range(4, self.N + 1):
                l_tab_T[l_idx] = (1, i)
                sl_tab = young.Tableau(zip(*l_tab_T))
                r_tab_T[r_idx] = (2, i)
                sr_tab = young.Tableau(zip(*r_tab_T))
                result[LRTableau(self.N, sl_tab, sr_tab)] -= coeff
        return result

    def ibp3b(
        self,
        coeff: Optional[utils.Fraction] = None,
    ) -> Count["LRTableau"]:
        if coeff is None:
            coeff = utils.Fraction(1)
        result = Counter()
        l_tab, r_tab = self.left_tableau, self.right_tableau
        l_tab_T, r_tab_T = list(zip(*l_tab)), list(zip(*r_tab))
        try:
            r_idx = r_tab_T.index((1, 3))
            l_idx = l_tab_T.index((2, 3))
        except ValueError:
            pass
        else:
            for i in range(4, self.N + 1):
                r_tab_T[r_idx] = (1, i)
                sr_tab = young.Tableau(zip(*r_tab_T))
                l_tab_T[l_idx] = (2, i)
                sl_tab = young.Tableau(zip(*l_tab_T))
                result[LRTableau(self.N, sl_tab, sr_tab)] -= coeff
        return result

    def ibp4(
        self,
        coeff: Optional[utils.Fraction] = None,
    ) -> Count["LRTableau"]:
        if coeff is None:
            coeff = utils.Fraction(1)
        result = Counter()
        l_tab, r_tab = self.left_tableau, self.right_tableau
        l_tab_T, r_tab_T = list(zip(*l_tab)), list(zip(*r_tab))
        try:
            l_idx = l_tab_T.index((2, 3))
            r_idx = r_tab_T.index((2, 3))
        except ValueError:
            pass
        else:
            for i in range(2, self.N + 1):
                for j in range(max(4, i + 1), self.N + 1):
                    l_tab_T[l_idx] = (i, j)
                    sl_tab = young.Tableau(zip(*l_tab_T))
                    r_tab_T[r_idx] = (i, j)
                    sr_tab = young.Tableau(zip(*r_tab_T))
                    result[LRTableau(self.N, sl_tab, sr_tab)] -= coeff
        return result

    def render_eps(self, prefix: str = "") -> str:
        def _index(index, counter):
            counter[index] += 1
            return counter[index]

        def _render_index(index, index_num):
            if prefix:
                return f"{prefix}_{index}_{_index(index,index_num)}"
            return f"{index}_{_index(index,index_num)}"

        index_num = Counter()
        undotted = [
            "eps(" + ",".join(_render_index(index, index_num) for index in column) + ")"
            for column in self.left_tableau.transposed()
        ]
        index_num.clear()
        dotted = [
            "eps("
            + ",".join(
                f"{_render_index(index,index_num)}{autoeft.epsilon_dot_symbol}"
                for index in column
            )
            + ")"
            for column in self.right_tableau.transposed()
        ]
        return "*".join(undotted + dotted)

    def render_eps_tex(self, lorentz_group: symmetry_groups.LorentzGroup) -> str:
        def _index(index, counter):
            counter[index] += 1
            return counter[index]

        def _render_index(index, index_num, dot: bool):
            if dot:
                return lorentz_group.get_r_idx(index, _index(index, index_num))
            return lorentz_group.get_l_idx(index, _index(index, index_num))

        index_num = Counter()
        undotted = [
            r"\epsilon^{"
            + " ".join(_render_index(index, index_num, dot=False) for index in column)
            + "}"
            for column in self.left_tableau.transposed()
        ]
        index_num.clear()
        dotted = [
            r"\epsilon_{"
            + " ".join(f"{_render_index(index,index_num,dot=True)}" for index in column)
            + "}"
            for column in self.right_tableau.transposed()
        ]
        return " ".join(undotted + dotted)


# SU(N) structures


class SUNTensor(utils.Vector):
    def __form__(self):
        return " + ".join(f"({coeff})*({form(tab)})" for tab, coeff in self.items())

    def symmetrize(self, symmetrizer: Counter, key: Callable) -> "SUNTensor":
        res = sum(
            (coeff * tab.symmetrized(symmetrizer, key) for tab, coeff in self.items()),
            start=SUNTensor(),
        )
        self.clear()
        self.update(res)
        return self

    def reduction(
        self,
        basis: List["SUNTableau"],
        group: symmetry_groups.SUNGroup,
        representations: List[reprs.SUNRepr],
        inverse_gram: Optional[sage.matrix.constructor.Matrix] = None,
    ) -> "SUNTensor":
        if not self:
            return SUNTensor()
        dim = len(basis)
        if not inverse_gram:
            inverse_gram = calculate_gram_matrix(
                basis,
                group,
                representations,
            ).inverse()
        it = (form(tab.symmetrize_sun_indices(representations)) for tab in basis)
        v = contract.contract_basis(form(self), it, group.N, dim)
        coeffs = sage.modules.free_module_element.vector(QQ, dim, v) * inverse_gram
        return SUNTensor(dict(zip(basis, coeffs)))


class SUNTableau(young.Tableau):
    def __form__(self) -> str:
        return "*".join(
            "e_(" + ",".join(f"F{index[0]}I{index[1]}" for index in column) + ")"
            for column in self.transposed()
        )

    def permuted(self, perm: Tuple[int, ...], key: Callable) -> "SUNTableau":
        perm_map = dict(zip(sorted(perm), perm))
        return SUNTableau((key(perm_map, index) for index in row) for row in self)

    def symmetrized(self, symmetrizer: Counter, key: Callable) -> SUNTensor:
        return sum(
            (
                SUNTensor({self.permuted(perm, key): coeff})
                for perm, coeff in symmetrizer.items()
            ),
            start=SUNTensor(),
        )

    def symmetrize_sun_indices(self, representations: List[reprs.SUNRepr]) -> SUNTensor:
        tensor = SUNTensor({self: 1})
        for field_num, field_repr in enumerate(representations, start=1):
            normal_tableau = field_repr.partition.normal_tableau()
            young_operator = sage.combinat.symmetric_group_algebra.e_hat(normal_tableau)
            algebra = young_operator.parent()
            young_operator_antipode = algebra.antipode(young_operator)
            symmetrizer = algebra.left_action_product(
                young_operator_antipode,
                young_operator,
            )
            index_symmetrizer = Counter(
                {
                    perm: utils.Fraction(
                        int(coeff.numerator()),
                        int(coeff.denominator()),
                    )
                    for perm, coeff in symmetrizer
                },
            )
            tensor.symmetrize(
                index_symmetrizer,
                key=lambda x, y: (y[0], x.get(y[1], y[1])) if field_num == y[0] else y,
            )
        return tensor

    def render_eps(self, prefix: str = "") -> str:
        def _render_index(index, index_num):
            if prefix:
                return f"{prefix}_{index}_{index_num}"
            return f"{index}_{index_num}"

        return "*".join(
            "eps("
            + ",".join(_render_index(index[0], index[1]) for index in column)
            + ")"
            for column in self.transposed()
        )

    def render_tableau(
        self,
        op_type: classes.Type,
        group_name: str,
        extra_sign: int = 1,
    ) -> str:
        sign = "+" if extra_sign == 1 else "-"
        tensor = self.render_eps() or "1"
        fields = []
        field_num = 0
        for f, m in op_type:
            for i in range(1, m + 1):
                field = f.name
                n = sum(f.representations[group_name].partition)
                if n:
                    field_indices = ",".join(
                        f"{field_num+i}_{j}" for j in range(1, n + 1)
                    )
                    field += f"({field_indices})"
                fields.append(field)
            field_num += m
        fields = "*".join(fields)
        return sign + tensor + " * " + fields

    def render_eps_tex(self, sun_group: symmetry_groups.SUNGroup) -> str:
        return " ".join(
            r"\epsilon^{"
            + " ".join(sun_group.get_idx(*index) for index in column)
            + "}"
            for column in self.transposed()
        )

    def render_tableau_tex(
        self,
        op_type: classes.Type,
        sun_group: symmetry_groups.SUNGroup,
        extra_sign: int = 1,
    ) -> str:
        sign = "+" if extra_sign == 1 else "-"
        tensor = self.render_eps_tex(sun_group) or "1"
        fields = []
        field_num = 0
        for f, m in op_type:
            for i in range(1, m + 1):
                field = f.tex
                n = sum(f.representations[sun_group.name].partition)
                if n:
                    field_indices = " ".join(
                        sun_group.get_idx(field_num + i, j) for j in range(1, n + 1)
                    )
                    field = f"{{{field}}}_{{{field_indices}}}"
                fields.append(field)
            field_num += m
        fields = r"\,".join(fields)
        return rf"{sign} \, {tensor} \; {fields}"

    @classmethod
    def parse_eps(cls, tensor: str) -> "SUNTableau":
        tab = []
        for eps in tensor.split("*"):
            if match := eps_pattern.match(eps):
                tab.append(
                    tuple(map(int, idx.split("_"))) for idx in match[1].split(",")
                )
            else:
                errmsg = (
                    "Invalid SU(N) tensor."
                    "\n"
                    "Tensors must be a product of 'eps(...)'."
                    "\n"
                    f"{eps}"
                )
                raise ValueError(errmsg)
        return cls(zip(*tab))

    @classmethod
    def parse_tableau(cls, invariant: str) -> Tuple["SUNTableau", int]:
        signed_tensor = invariant.split(" * ")[0]
        if signed_tensor == "+1":
            return SUNTableau(), 1
        if signed_tensor == "-1":
            return SUNTableau(), -1
        sign = signed_tensor[0]
        tensor = signed_tensor[1:]
        tab = cls.parse_eps(tensor)
        sign_map = {"+": 1, "-": -1}
        return tab, sign_map[sign]

    @classmethod
    def fill_tableau(cls, field_repr: reprs.SUNRepr, field_num: int) -> "SUNTableau":
        """Construct tableau of given shape with field indices in canonical ordering.

        The tableau is filled with increasing indices
        from left to right and top to bottom.
        """
        index_num = 0
        return cls(
            ((field_num, index_num := index_num + 1) for _ in range(row_len))
            for row_len in field_repr
        )


# combined structures


class CombinedTensor(utils.Vector):
    def render_term(self, op_type: classes.Type) -> str:
        def _render_tableaux(coeff: int, tableaux: CombinedTableaux):
            rendered, sign = tableaux.render_tableaux(op_type)
            return f"({coeff*sign}) * {rendered}"

        return " + ".join(
            _render_tableaux(coeff, tableaux) for tableaux, coeff in self.items()
        )


class CombinedTableaux(dict):
    def __hash__(self):
        return hash(frozenset(self.items()))

    def render_eps(self) -> str:
        return "*".join(
            tableau.render_eps(group) for group, tableau in self.items() if tableau
        )

    def render_tableaux(
        self,
        op_type: classes.Type,
    ) -> Tuple[str, int]:
        tensor = self.render_eps()
        lr_tab, lr_sign = None, 1
        fields = []
        field_num = 0
        for f, m in op_type:
            for i in range(1, m + 1):
                nD = 0
                field_indices = {}
                for group, tableau in self.items():
                    group_indices = []
                    if isinstance(tableau, LorentzTableau):
                        if not lr_tab:
                            lr_tab, lr_sign = tableau.to_lr_tableau()
                        n_undotted = lr_tab.undotted_indices[field_num + i]
                        n_dotted = lr_tab.dotted_indices[field_num + i]
                        nD = min(n_undotted, n_dotted)
                        undotted = [
                            f"{group}_{field_num+i}_{j}"
                            for j in range(1, n_undotted + 1)
                        ]
                        dotted = [
                            f"{group}_{field_num+i}_{j}{autoeft.epsilon_dot_symbol}"
                            for j in range(1, n_dotted + 1)
                        ]
                        group_indices = undotted + dotted
                    else:
                        n = sum(f.representations[group].partition)
                        group_indices = [
                            f"{group}_{field_num+i}_{j}" for j in range(1, n + 1)
                        ]
                    field_indices[group] = group_indices
                building_block = ""
                if nD == 1:
                    building_block = f"(D {f.name})"
                elif nD > 1:
                    building_block = f"(D^{nD} {f.name})"
                else:
                    building_block = f.name
                indices = ""
                indices = ";".join(
                    ",".join(group_indices)
                    for group_indices in field_indices.values()
                    if group_indices
                )
                if indices:
                    fields.append(f"{building_block}({indices})")
                else:
                    fields.append(building_block)
            field_num += m
        fields = "*".join(fields)
        return tensor + " * " + fields, lr_sign


# misc


def ibp_reduction(
    lr_tableau: LRTableau,
    coeff: Optional[utils.Fraction] = None,
) -> Count[LRTableau]:
    if coeff is None:
        coeff = utils.Fraction(1)
    result = Counter()
    lr_tab, lr_sign = lr_tableau.normalized()
    if not lr_sign * coeff:
        return result
    if ibp := lr_tab.ibp1():
        for tab, factor in ibp.items():
            result.update(ibp_reduction(tab, lr_sign * coeff * factor))
        return result
    if ibp := lr_tab.ibp2a():
        for tab, factor in ibp.items():
            result.update(ibp_reduction(tab, lr_sign * coeff * factor))
        return result
    if ibp := lr_tab.ibp2b():
        for tab, factor in ibp.items():
            result.update(ibp_reduction(tab, lr_sign * coeff * factor))
        return result
    if ibp := lr_tab.ibp3a():
        for tab, factor in ibp.items():
            result.update(ibp_reduction(tab, lr_sign * coeff * factor))
        return result
    if ibp := lr_tab.ibp3b():
        for tab, factor in ibp.items():
            result.update(ibp_reduction(tab, lr_sign * coeff * factor))
        return result
    if ibp := lr_tab.ibp4():
        for tab, factor in ibp.items():
            result.update(ibp_reduction(tab, lr_sign * coeff * factor))
        return result
    if schouten := lr_tab.schouten():
        for tab, factor in schouten.items():
            result.update(ibp_reduction(tab, lr_sign * coeff * factor))
        return result
    result[lr_tab] = lr_sign * coeff
    return result


def calculate_gram_matrix(
    basis: List[SUNTableau],
    group: symmetry_groups.SUNGroup,
    representations: List[reprs.SUNRepr],
    threads: Optional[int] = None,
) -> sage.matrix.constructor.Matrix:
    dim = len(basis)
    form_sbasis = [form(tab.symmetrize_sun_indices(representations)) for tab in basis]
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=threads or os.cpu_count(),
    ) as executor:
        matrix = list(
            executor.map(
                contract.contract_basis,
                (form(tab) for tab in basis),
                (form_sbasis[i:] for i in range(len(basis))),
                itertools.repeat(group.N),
                (dim - i for i in range(len(basis))),
            ),
        )
        return sage.matrix.constructor.matrix(
            QQ,
            dim,
            dim,
            lambda i, j: matrix[i][j - i] if i <= j else matrix[j][i - j],
        )



class MissingResourcesError(Exception):
    pass


class MissingRequirementError(Exception):
    pass


class RequirementVersionError(Exception):
    pass


class NoContractionException(Exception):
    pass


class InvalidRangeException(Exception):
    pass


class BasisNotFoundError(Exception):
    pass



import argparse
import datetime
from pathlib import Path
from typing import Optional
from typing import Tuple

import pydantic
import yaml

import autoeft
import autoeft.base.basis as eft_basis
import autoeft.base.model as ir_model
from autoeft import exceptions
from autoeft import utils


class MetaData(pydantic.BaseModel):
    timestamp: datetime.datetime
    eft_name: str
    dimension: int
    args: dict

    @pydantic.validator("args")
    def validate_args(cls, v):
        for k in "model", "output", "generators":
            if k in v:
                v[k] = Path(v[k])
        return v


class BasisFile:
    basis_path: Path
    model_path: Path
    extra_path: Path

    def __init__(self, basis_path: Path):
        self.basis_path = basis_path
        self.model_path = basis_path / "model.json"
        self.extra_path = basis_path / ".autoeft"

    def __bool__(self):
        return (
            self.basis_path.is_dir()
            and self.model_path.is_file()
            and self.extra_path.is_file()
        )

    def get_model(self) -> ir_model.Model:
        try:
            return ir_model.Model.parse_json(self.model_path.read_text())
        except FileNotFoundError as e:
            errmsg = f"No '{self.model_path.name}' in basis directory."
            raise exceptions.BasisNotFoundError(errmsg) from e

    def set_model(self, model: ir_model.Model):
        self.basis_path.mkdir(parents=True, exist_ok=True)
        data = model.json(indent=2)
        self.model_path.write_text(data)

    def get_extras(self):
        try:
            return [
                MetaData(**data) for data in yaml.safe_load(self.extra_path.read_text())
            ]
        except FileNotFoundError as e:
            errmsg = f"No '{self.extra_path.name}' in basis directory."
            raise exceptions.BasisNotFoundError(errmsg) from e
        except pydantic.ValidationError as e:
            errmsg = f"{e}\nCannot parse '{self.extra_path.name}'."
            raise exceptions.BasisNotFoundError(errmsg) from e

    def get_extra(self):
        return self.get_extras()[-1]

    def set_extra(
        self,
        eft_name: str,
        dimension: int,
        args: argparse.Namespace,
    ):
        exclude = {"main_func", "quiet", "verbose", "dry_run", "name"}
        args_dict = {k: v for k, v in vars(args).items() if k not in exclude}
        meta = MetaData.construct(
            timestamp=datetime.datetime.now().isoformat(timespec="seconds"),
            eft_name=eft_name,
            dimension=dimension,
            args=args_dict,
        )
        self.basis_path.mkdir(parents=True, exist_ok=True)
        with self.extra_path.open("a") as extra_file:
            extra_file.write(
                yaml.dump([meta.dict()], sort_keys=False, Dumper=utils.YamlDumper),
            )

    def add_operator_info(self, operator_info: eft_basis.OperatorInfoPermutation):
        self.basis_path.mkdir(parents=True, exist_ok=True)
        operator_path = self.basis_path / operator_info.path()
        operator_path.parent.mkdir(parents=True, exist_ok=True)
        data = (
            f"# '{operator_path.name}' generated by "
            f"{autoeft.program} {autoeft.version}\n"
        )
        data += operator_info.yaml()
        tmp_operator_path = operator_path.with_suffix(".tmp")
        tmp_operator_path.write_text(data)
        tmp_operator_path.rename(operator_path)

    def get_operator_info(self, rel_path: Path) -> eft_basis.OperatorInfoPermutation:
        model = self.get_model()
        operator_path = self.basis_path / rel_path
        try:
            return eft_basis.OperatorInfoPermutation(
                model=model,
                **yaml.safe_load(operator_path.read_text()),
            )
        except FileNotFoundError as e:
            errmsg = f"No '{rel_path}' in basis directory."
            raise exceptions.BasisNotFoundError(errmsg) from e

    def get_operator_raw_numbers(self, rel_path: Path) -> Optional[Tuple[int, int]]:
        operator_path = self.basis_path / rel_path
        if operator_path.is_file():
            data = yaml.safe_load(operator_path.read_text())
            return data["n_terms"], data["n_operators"]
        return None

    def get_basis(self) -> eft_basis.Basis:
        def _sort(op_info: eft_basis.OperatorInfo):
            family = tuple(-nh for nh in op_info.op_type.family)
            op_type = tuple((f.name, m) for f, m in op_info.op_type)
            return op_info.N, (-op_info.nl, -op_info.nr), family, op_type

        model = self.get_model()
        operators = {}
        for operator_path in self.basis_path.rglob("*.yml"):
            op_info = eft_basis.OperatorInfoPermutation(
                model=model,
                **yaml.safe_load(operator_path.read_text()),
            )
            operators[repr(op_info)] = op_info
        operators_sorted = sorted(operators.items(), key=lambda item: _sort(item[1]))
        operator_basis = eft_basis.Basis(model)
        operator_basis.update(operators_sorted)
        return operator_basis



import re

import pydantic

from autoeft import utils
from autoeft.combinatorics import young


class Repr(pydantic.BaseModel):
    def __rich__(self):
        return str(self)

    def conjugated(self) -> "Repr":
        raise NotImplementedError

    def is_real(self) -> bool:
        return self == self.conjugated()

    def is_complex(self) -> bool:
        return self != self.conjugated()

    class Config:
        extra = pydantic.Extra.forbid


class LorentzRepr(Repr):
    """A representation of the Lorentz group.

    Characterized by the helicity h = j_r - j_l
    for a massless field in the (j_l, j_r) representation.
    """

    helicity: utils.Fraction

    def __str__(self):
        return str(self.helicity)

    def __rich__(self):
        return self.helicity

    def __bool__(self):
        return bool(self.helicity)

    @pydantic.validator("helicity")
    def parse_helicity(cls, v):
        if v not in (
            utils.Fraction(-2),
            utils.Fraction(-1),
            utils.Fraction(-1, 2),
            utils.Fraction(0),
            utils.Fraction(1, 2),
            utils.Fraction(1),
            utils.Fraction(2),
        ):
            errmsg = f"helicity can only be 0, +-1/2, +-1, or +-2, not {v}"
            raise ValueError(errmsg)
        return v

    def conjugated(self) -> "LorentzRepr":
        """Construct the hermitian conjugate representation."""
        return self.copy(update={"helicity": -self.helicity})


class SUNRepr(Repr):
    """A representation of an SU(N) group.

    Characterized by a SU(N) Young diagram.
    """

    N: pydantic.PositiveInt
    partition: young.Partition

    def __str__(self):
        return str(self.partition)

    def __rich__(self):
        return self.partition

    def __bool__(self):
        return bool(self.partition)

    def __iter__(self):
        return iter(self.partition)

    @pydantic.validator("partition", pre=True)
    def parse_partition(cls, v, values):
        if isinstance(v, (list, young.Partition)):
            if "N" in values and len(v) >= (N := values["N"]):
                errmsg = f"{v} is not a valid SU({N}) representation"
                raise ValueError(errmsg)
            return v
        if isinstance(v, tuple):
            dynkin = v
            if "N" in values and (len(dynkin) + 1) != (N := values["N"]):
                errmsg = f"{v} is not a valid SU({N}) representation"
                raise ValueError(errmsg)
            return young.Partition.from_dynkin(dynkin)
        if isinstance(v, str) and (result := re.match(r"^\((.*)\)$", v)):
            dynkin = list(map(int, result[1].split(",")))
            if "N" in values and (len(dynkin) + 1) != (N := values["N"]):
                errmsg = f"{v} is not a valid SU({N}) representation"
                raise ValueError(errmsg)
            return young.Partition.from_dynkin(dynkin)
        errmsg = f"{v} is not a valid SU(N) representation"
        raise ValueError(errmsg)

    def conjugated(self) -> "SUNRepr":
        """Construct the hermitian conjugate representation.

        This inverts the order of the Dynkin labels of the partition.
        """
        conj_part = young.Partition.from_dynkin(
            list(reversed(self.partition.to_dynkin(self.N))),
        )
        return self.copy(update={"N": self.N, "partition": conj_part})

    def is_real(self):
        dynkin = self.partition.to_dynkin(self.N)
        if dynkin == list(reversed(dynkin)):
            if (self.N + 2) % 4 == 0:
                return dynkin[int(self.N / 2 - 1)] % 2 == 0
            return True
        return False

    def is_pseudo_real(self) -> bool:
        return not self.is_complex() and not self.is_real()


class U1Repr(Repr):
    """A representation of a U(1) group.

    Characterized by a U(1) charge.
    """

    charge: utils.Fraction

    def __str__(self):
        return str(self.charge)

    def __rich__(self):
        return self.charge

    def __bool__(self):
        return bool(self.charge)

    def conjugated(self) -> "U1Repr":
        """Construct the hermitian conjugate representation."""
        return self.copy(update={"charge": -self.charge})



import concurrent.futures
import itertools
import os
from collections import Counter
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

import sage.all
import sage.matrix.constructor
from sage.rings.rational_field import QQ

import autoeft.base.classifications as classes
import autoeft.base.groups as symmetry_groups
from autoeft import utils
from autoeft.base import tensors
from autoeft.combinatorics import generators
from autoeft.combinatorics import young


class GroupSymmetrizer(NamedTuple):
    group_name: str
    fields_info: Tuple[Tuple[str, young.Partition, int], ...]


def construct_field_symmetrizers(
    group_symmetrizer: GroupSymmetrizer,
    op_type: classes.Type,
) -> Dict[str, Counter]:
    field_numbers = op_type.field_numbers()
    field_symmetrizers = {}
    for field_name, partition, basis_element in group_symmetrizer.fields_info:
        symmetrizer = generators.get_young_basis(partition)[basis_element]
        field_symmetrizers[field_name] = Counter(
            {
                tuple(
                    field_numbers[field_name][entry - 1] for entry in perm
                ): utils.Fraction(int(coeff.numerator()), int(coeff.denominator()))
                for perm, coeff in symmetrizer
            },
        )
    return field_symmetrizers


# Lorentz structures


def construct_symmetrized_lorentz_tensors(
    group_symmetrizer: GroupSymmetrizer,
    op_type: classes.Type,
    tableaux: List[tensors.LorentzTableau],
) -> Iterator[tensors.LorentzTensor]:
    field_symmetrizers = construct_field_symmetrizers(group_symmetrizer, op_type)
    for tableau in tableaux:
        tensor = tensors.LorentzTensor({tableau: 1})
        for field_symmetrizer in field_symmetrizers.values():
            tensor.symmetrize(field_symmetrizer)
        yield tensor.reduce()


def construct_lorentz_matrix(
    op_type: classes.Type,
    group_symmetrizer: GroupSymmetrizer,
    basis: List[tensors.LorentzTableau],
) -> sage.matrix.constructor.Matrix:
    dim = len(basis)
    symmetrized_tensors = construct_symmetrized_lorentz_tensors(
        group_symmetrizer,
        op_type,
        basis,
    )
    matrix = []
    for symmetrized_tensor in symmetrized_tensors:
        tensor = symmetrized_tensor.reduction(
            basis,
            op_type.family.N,
            op_type.family.nr,
        )
        matrix.append([tensor[tab] for tab in basis])
    return sage.matrix.constructor.matrix(QQ, dim, dim, matrix)


# SU(N) structures


def construct_symmetrized_sun_tensors(
    group_symmetrizer: GroupSymmetrizer,
    op_type: classes.Type,
    tableaux: List[tensors.SUNTableau],
) -> Iterator[tensors.SUNTensor]:
    field_symmetrizers = construct_field_symmetrizers(group_symmetrizer, op_type)
    for tableau in tableaux:
        tensor = tensors.SUNTensor({tableau: 1})
        for field_symmetrizer in field_symmetrizers.values():
            tensor.symmetrize(
                field_symmetrizer,
                key=lambda x, y: (x.get(y[0], y[0]), y[1]),
            )
        yield tensor.reduce()


def sun_tensor_reduction(symmetrized_tensor, basis, group, representations, igram):
    tensor = symmetrized_tensor.reduction(
        basis,
        group,
        representations,
        igram,
    )
    return [tensor[tab] for tab in basis]


def construct_sun_matrix(
    op_type: classes.Type,
    group_symmetrizer: GroupSymmetrizer,
    basis: List[tensors.SUNTableau],
    group: symmetry_groups.SUNGroup,
    igram_cache: Dict[str, sage.matrix.constructor.Matrix],
    threads: Optional[int] = None,
) -> sage.matrix.constructor.Matrix:
    dim = len(basis)
    representations = op_type.sun_content(group)
    symmetrized_tensors = construct_symmetrized_sun_tensors(
        group_symmetrizer,
        op_type,
        basis,
    )
    if group.name not in igram_cache:
        igram_cache[group.name] = tensors.calculate_gram_matrix(
            basis,
            group,
            representations,
            threads,
        ).inverse()
    igram = igram_cache[group.name]
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=threads or os.cpu_count(),
    ) as executor:
        matrix = executor.map(
            sun_tensor_reduction,
            symmetrized_tensors,
            itertools.repeat(basis),
            itertools.repeat(group),
            itertools.repeat(representations),
            itertools.repeat(igram),
        )
        return sage.matrix.constructor.matrix(QQ, dim, dim, matrix)



import argparse
import os
import sys
from pathlib import Path
from typing import List
from typing import Optional

import autoeft
import autoeft.base.model as model_entry
import autoeft.combinatorics.generators as generators_entry
import autoeft.construction.main as construction_entry
import autoeft.io.check as check_entry
import autoeft.io.count as count_entry
import autoeft.tex.main as latex_entry

print = autoeft.print  # noqa: A001


def create_chk_subparser(subparsers):
    parser = subparsers.add_parser(
        "check",
        add_help=False,
        help="check functionality of this installation",
    )
    parser.set_defaults(main_func=check_entry.main)
    return parser


def create_spl_subparser(subparsers):
    parser = subparsers.add_parser(
        "sample-model",
        add_help=False,
        help="produce a sample model file",
    )
    parser.set_defaults(main_func=model_entry.main, quiet=True)
    return parser


def create_con_subparser(subparsers):
    desc = (
        "Construct an operator basis for a given IR model and mass dimension. "
        "The operator basis is a minimal set of non-redundant operators. "
        "In order to generate an operator basis,"
        " provide a model file and the desired mass dimension."
    )
    parser = subparsers.add_parser(
        "construct",
        aliases=["c"],
        description=desc,
        help="construct an operator basis for a given model and dimension",
    )
    parser.add_argument("model", type=Path, help="model file")
    parser.add_argument("dimension", type=int, help="mass dimension")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="show intermediate construction output",
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=os.cpu_count(),
        help="set number of threads starting FORM (default: %(default)s)",
    )
    parser.add_argument(
        "-n",
        "--name",
        type=str,
        help="set the EFT name (default: <model>-eft)",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="PATH",
        type=Path,
        default=Path().cwd() / "efts",
        help="set output path (default: %(default)s)",
    )
    parser.add_argument(
        "-s",
        "--select",
        action="append",
        type=str,
        help="only construct selected types (default: All)",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        action="append",
        type=str,
        help="do not construct ignored types (default: %(default)s)",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="only list the types that match the current selection",
    )
    parser.add_argument(
        "--generators",
        metavar="PATH",
        type=Path,
        default=Path().cwd() / "gens",
        help="set generators path (default: %(default)s)",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="overwrite existing output files",
    )
    parser.add_argument(
        "--no_hc",
        action="store_true",
        help="prevent the explicit construction of conjugate operator types",
    )
    parser.set_defaults(main_func=construction_entry.main)
    return parser


def create_cnt_subparser(subparsers):
    desc = (
        "Count the number of families, types, terms, and operators for a given basis."
    )
    parser = subparsers.add_parser(
        "count",
        description=desc,
        help="count invariant classes for a given basis",
    )
    parser.add_argument("basis", type=Path, help="basis directory")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="show model details and operator type numbers",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        type=Path,
        default=Path().cwd() / "counts.yml",
        help="write the numbers to FILE (default: %(default)s)",
    )
    parser.add_argument(
        "-s",
        "--select",
        action="append",
        type=str,
        help="only include selected types (default: All)",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        action="append",
        type=str,
        help="do not include ignored types (default: %(default)s)",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="only list the types that match the current selection",
    )
    parser.add_argument(
        "--no_hc",
        action="store_true",
        help="prevent the implicit counting of conjugate operator types",
    )
    parser.set_defaults(main_func=count_entry.main)
    return parser


def create_tex_subparser(subparsers):
    desc = (
        "Generate TeX files for a given basis."
        " The TeX files represent all the information encoded in the operator files"
        " as LaTeX markup."
        " The resulting files compose a valid LaTeX document"
        " that can be compiled to a single PDF file."
    )
    parser = subparsers.add_parser(
        "latex",
        aliases=["l"],
        description=desc,
        help="generate TeX files for a given basis",
    )
    parser.add_argument("basis", type=Path, help="basis directory")
    parser.add_argument(
        "-c",
        "--compile",
        type=str,
        metavar="COMMAND",
        help="compile TeX files using `COMMAND`",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="PATH",
        type=Path,
        default=Path().cwd() / "tex",
        help="set output path (default: %(default)s)",
    )
    parser.add_argument(
        "-s",
        "--select",
        action="append",
        type=str,
        help="only include selected types (default: All)",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        action="append",
        type=str,
        help="do not include ignored types (default: %(default)s)",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="only list the types that match the current selection",
    )
    parser.set_defaults(main_func=latex_entry.main)
    return parser


def create_gen_subparser(subparsers):
    desc = (
        "View or create generator matrices of"
        " symmetric group representations. "
        "These matrices are used to generate"
        " the irreducible representation matrices. "
        "The generators can be constructed"
        " for the whole symmetric group"
        " or specific irreducible representations."
    )
    parser = subparsers.add_parser(
        "generators",
        aliases=["g"],
        description=desc,
        help="view or create generators of the symmetric group",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="PATH",
        type=Path,
        default=Path().cwd() / "gens",
        help="set output path (default: %(default)s)",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-S",
        dest="N",
        type=int,
        help="construct all generators for the symmetric group of degree `N`",
    )
    group.add_argument(
        "-P",
        dest="partition",
        metavar="p",
        nargs="+",
        type=int,
        help=(
            "construct all generators for the irreducible representation "
            "given by the partition as a non-increasing list of integers `p`"
        ),
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="overwrite existing output files",
    )
    parser.set_defaults(main_func=generators_entry.main)
    return parser


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=autoeft.__summary__,
        epilog=str(autoeft.disclaimer),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=str(autoeft.version),
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="suppress all output",
    )
    subparsers = parser.add_subparsers(title="commands")
    create_chk_subparser(subparsers)
    create_spl_subparser(subparsers)
    create_con_subparser(subparsers)
    create_cnt_subparser(subparsers)
    create_tex_subparser(subparsers)
    create_gen_subparser(subparsers)
    return parser


def main(cli_args: Optional[List[str]] = None) -> int:
    print(*sys.argv, dest="log")
    parser = create_parser()
    args = parser.parse_args(args=cli_args)
    if not hasattr(args, "main_func"):
        parser.print_help()
        raise SystemExit
    autoeft.console.quiet = args.quiet
    print(autoeft.logo)
    print(autoeft.disclaimer)
    print()
    return args.main_func(args)



from typing import Optional
from typing import Tuple

import pydantic
import rich.text

from autoeft import utils

GroupName = pydantic.constr(regex=r"^[a-zA-Z][a-zA-Z0-9()]*[a-zA-Z0-9()]$|^[a-zA-Z]$")


class Group(pydantic.BaseModel):
    name: GroupName
    tex: Optional[str] = None

    def __str__(self):
        return self.name

    def __tex__(self):
        return self.tex

    def __rich__(self):
        return rich.text.Text(str(self), "bold")

    @pydantic.validator("tex", always=True)
    def parse_tex(cls, v, values):
        if "name" in values:
            name = values["name"]
            return v or rf"\mathtt{{{name}}}"
        return v

    class Config:
        frozen = True
        extra = pydantic.Extra.forbid


class LorentzGroup(Group):
    """The universal covering group SL(2,CC) of the restricted Lorentz symmetry."""

    name: str = "Lorentz"
    indices: Tuple[str, ...] = (
        r"\alpha",
        r"\beta",
        r"\gamma",
        r"\delta",
        r"\varepsilon",
        r"\zeta",
        r"\eta",
        r"\theta",
        r"\iota",
        r"\kappa",
        r"\lambda",
    )

    def get_l_idx(self, i: int, j: int):
        """Return an undotted spinor index."""
        p = r"\prime"
        i = i - 1
        m = len(self.indices)
        idx = f"{self.indices[i % m]}_{{{j}}}"
        if repeat := i // m:
            return f"{idx}^{{{p*repeat}}}"
        return idx

    def get_r_idx(self, i: int, j: int):
        """Return a dotted spinor index."""
        p = r"\prime"
        i = i - 1
        m = len(self.indices)
        idx = rf"\dot{self.indices[i % m]}_{{{j}}}"
        if repeat := i // m:
            return f"{idx}^{{{p*repeat}}}"
        return idx


class SUNGroup(Group):
    """An SU(N) internal symmetry."""

    N: pydantic.conint(gt=1)
    indices: Tuple[str, ...] = tuple("abcdefghijk")

    def get_idx(self, i: int, j: int):
        """Return a symmetry/gauge index."""
        p = r"\prime"
        i = i - 1
        m = len(self.indices)
        idx = f"{self.indices[i % m]}_{{{j}}}"
        if repeat := i // m:
            return f"{idx}^{{{p*repeat}}}"
        return idx


class U1Group(Group):
    """A U(1) symmetry/gauge group."""

    violation: utils.Fraction = utils.Fraction(0)
    residual: utils.Fraction = utils.Fraction(0)



import importlib.resources
import sys
from pathlib import Path

import yaml

import autoeft
import autoeft.base.basis as eft_basis
import autoeft.base.model as ir_model
import autoeft.construction.main as con_main
from autoeft import exceptions

print = autoeft.print  # noqa: A001


default_module = __name__.rsplit(".", 1)[0]


def get_resource_basis(
    basis_path: Path,
) -> eft_basis.Basis:
    model_path = Path("model.json")
    try:
        model = ir_model.Model.parse_json(
            importlib.resources.files(default_module)
            .joinpath(basis_path, model_path)
            .read_text(),
        )
    except FileNotFoundError as e:
        errmsg = f"No '{model_path.name}' in basis directory."
        raise exceptions.BasisNotFoundError(errmsg) from e

    def traverse(anchor):
        for resource in anchor.iterdir():
            if resource.is_file() and resource.name.endswith(".yml"):
                yield resource
            elif resource.is_dir() and resource.name != "__pychache__":
                yield from traverse(
                    importlib.resources.files(default_module).joinpath(
                        anchor,
                        resource,
                    ),
                )

    operators = {}
    for operator_path in traverse(
        importlib.resources.files(default_module).joinpath(basis_path),
    ):
        op_info = eft_basis.OperatorInfoPermutation(
            model=model,
            **yaml.safe_load(operator_path.read_text()),
        )
        operators[repr(op_info)] = op_info

    def _sort(op_info: eft_basis.OperatorInfo):
        family = tuple(-nh for nh in op_info.op_type.family)
        op_type = tuple((f.name, m) for f, m in op_info.op_type)
        return op_info.N, (-op_info.nl, -op_info.nr), family, op_type

    operators_sorted = sorted(operators.items(), key=lambda item: _sort(item[1]))
    operator_basis = eft_basis.Basis(model)
    operator_basis.update(operators_sorted)
    return operator_basis


def get_resource_basis_lagacy(
    basis_path: str,
) -> eft_basis.Basis:
    model_path = Path("model.json")
    try:
        model = ir_model.Model.parse_json(
            importlib.resources.read_text(f"{default_module}.{basis_path}", model_path),
        )
    except FileNotFoundError as e:
        errmsg = f"No '{model_path.name}' in basis directory."
        raise exceptions.BasisNotFoundError(errmsg) from e

    def traverse(anchor):
        for resource in importlib.resources.contents(anchor):
            if resource.endswith(".yml"):
                yield anchor, resource
            elif "." not in resource and resource not in {
                "__pycache__",
                "__init__.py",
            }:
                yield from traverse(f"{anchor}.{resource}")

    operators = {}
    for operator_module, operator_path in traverse(f"{default_module}.{basis_path}"):
        op_info = eft_basis.OperatorInfoPermutation(
            model=model,
            **yaml.safe_load(
                importlib.resources.read_text(operator_module, operator_path),
            ),
        )
        operators[repr(op_info)] = op_info

    def _sort(op_info: eft_basis.OperatorInfo):
        family = tuple(-nh for nh in op_info.op_type.family)
        op_type = tuple((f.name, m) for f, m in op_info.op_type)
        return op_info.N, (-op_info.nl, -op_info.nr), family, op_type

    operators_sorted = sorted(operators.items(), key=lambda item: _sort(item[1]))
    operator_basis = eft_basis.Basis(model)
    operator_basis.update(operators_sorted)
    return operator_basis


def main(args) -> int:
    try:
        if sys.version_info >= (3, 9):
            sample = (
                importlib.resources.files(autoeft.__program__)
                .joinpath("sm.yml")
                .read_text()
            )
        else:
            sample = importlib.resources.read_text(autoeft.__program__, "sm.yml")
    except FileNotFoundError:
        sample = "name: <sample-model>\nsymmetries: {}\nfields: {}\n"
    model = ir_model.Model(**yaml.safe_load(sample))
    print("Checking AutoEFT's functionality:")
    print("Constructing operator basis for [b]SMEFT[/b] @ [i]d=6[/i]")
    basis, _ = con_main.plain_main(model, 6, True, False)
    if sys.version_info >= (3, 9):
        resource_basis = get_resource_basis(
            Path(f"check_basis-{autoeft.version}".replace(".", "_")),
        )
    else:
        resource_basis = get_resource_basis_lagacy(
            f"check_basis-{autoeft.version}".replace(".", "_"),
        )
    print("Check against pre-constructed basis...")
    if basis != resource_basis:
        print("[b]Failure[/b] - found divergent bases!")
        return 1
    print("[b]Success[/b] - found complete agreement!")
    return 0



import datetime
import importlib
import shutil
import subprocess
import sys
from collections import Counter
from pathlib import Path
from typing import Counter as Count
from typing import Iterable
from typing import Iterator
from typing import Optional

import rich.markup
import yaml

import autoeft
import autoeft.base.model as ir_model
import autoeft.io.basis as io_basis
from autoeft import exceptions
from autoeft import utils
from autoeft.tex import tex

default_module = __name__.rsplit(".", 1)[0]

print = autoeft.print  # noqa: A001

try:
    if sys.version_info >= (3, 9):
        with importlib.resources.as_file(
            importlib.resources.files(default_module).joinpath("main.tex"),
        ) as file_path:
            texfile_path = file_path
    else:
        with importlib.resources.path(default_module, "main.tex") as file_path:
            texfile_path = file_path
except FileNotFoundError:
    texfile_path = None


def dry_run(
    basis,
    select: Optional[list] = None,
    ignore: Optional[list] = None,
):
    for op_info in basis.values():
        if select and not utils.type_matches(op_info.op_type, select):
            continue
        if ignore and utils.type_matches(op_info.op_type, ignore):
            continue
        yield op_info.op_type


def main(args) -> int:
    basis_path = args.basis
    basis_file = io_basis.BasisFile(basis_path)
    try:
        model = basis_file.get_model()
        basis = basis_file.get_basis()
        metadata = basis_file.get_extra()
    except exceptions.BasisNotFoundError as e:
        errmsg = f"{e}\nInvalid basis directory '{basis_file.basis_path}'."
        print(errmsg)
        return 1
    output_path = args.output.resolve()
    output_path /= f"{metadata.eft_name}"
    output_path /= ascii(metadata.dimension)
    output_path.mkdir(parents=True, exist_ok=True)
    operators_path = output_path / "operators"
    operators_path.mkdir(exist_ok=True)
    hc_flag = not metadata.args["no_hc"]
    try:
        select = (
            [yaml.safe_load(s.replace(":", ": ")) for s in args.select]
            if args.select
            else None
        )
        ignore = (
            [yaml.safe_load(i.replace(":", ": ")) for i in args.ignore]
            if args.ignore
            else None
        )
    except yaml.YAMLError as e:
        errmsg = f"{e}\nCould not parse select/ignore options."
        print(errmsg)
        return 1
    if args.dry_run:
        print(
            f"Dry run for "
            f"[b]{rich.markup.escape(model.name)}[/b] @ [i]d={metadata.dimension}[/i]",
        )
        print(f"[i]select[/i]: {select}")
        print(f"[i]ignore[/i]: {ignore}")
        print()
        print("[b]types:[/b]")
        for op_type in dry_run(
            basis,
            select,
            ignore,
        ):
            print(f"[bold white]{op_type}[/bold white]", dest="terminal")
            print(op_type, dest="log")
        return 0

    print(
        f"Generating TeX files for "
        f"[b]{rich.markup.escape(model.name)}[/b] @ [i]d={metadata.dimension}[/i]",
    )
    families = Counter()
    types = Counter()
    terms = Counter()
    operators = Counter()
    hilbert_series = Counter()
    tensor_product = rf"\[ \mathcal{{O}} \sim {tex(model.symmetries)} \]"
    operator_files = []
    for op_info in basis.values():
        if select and not utils.type_matches(op_info.op_type, select):
            continue
        if ignore and utils.type_matches(op_info.op_type, ignore):
            continue
        op_type = op_info.op_type
        family = op_type.family
        if hc_flag:
            tex_family = tex(family)
            tex_type = tex(op_type)
            hilbert_series[tex_type] = op_info.n_operators
            families[tex_family] = 1
            types[tex_family] += 1
            terms[tex_family] += op_info.n_terms
            operators[tex_family] += op_info.n_operators
        else:
            tex_family = tex(family)
            tex_type = tex(op_type)
            hilbert_series[tex_type] = op_info.n_operators
            hc_mod = 2 if op_type.is_complex() else 1
            if family.is_complex():
                tex_family += r" + \hc"
                families[tex_family] = 2
            else:
                families[tex_family] = 1
            types[tex_family] += 1 * hc_mod
            terms[tex_family] += op_info.n_terms * hc_mod
            operators[tex_family] += op_info.n_operators * hc_mod
        operator_path = operators_path / f"{op_type.path('tex')}"
        produce_texfile(
            operator_path,
            [
                rf"\subsection{{\texorpdfstring{{${tex(op_type)}$}}"
                f"{{{op_info.op_type!s}}}}}",
                tex(op_info),
            ],
        )
        operator_files.append(rf"\input{{operators/{operator_path.stem}}}")
    produce_texfile(
        output_path / "header.tex",
        tex_header(model, metadata.dimension, metadata.timestamp),
    )
    produce_texfile(
        output_path / "model.tex",
        [tex(model)],
    )
    produce_texfile(
        output_path / "numbers_table.tex",
        tex_table(families, types, terms, operators),
    )
    produce_texfile(
        output_path / "hilbert_series.tex",
        tex_series(hilbert_series, metadata.dimension, not hc_flag),
    )
    produce_texfile(
        output_path / "operators.tex",
        [tensor_product, *operator_files],
    )
    if texfile_path.is_file():
        if not (output_path / "main.tex").is_file():
            print(f"No 'main.tex' found, writing default file to {output_path}")
            shutil.copy(texfile_path, output_path)
        else:
            print("Existing 'main.tex' found")
    print(f"TeX files saved under {output_path}")
    if args.compile:
        print(f"[b]{args.compile}[/b]")
        result = subprocess.run(
            args.compile,
            cwd=output_path,
            shell=True,  # noqa: S602
            check=False,
        )
        return result.returncode
    return 0


def tex_header(model: ir_model.Model, dimension: int, timestamp: datetime.datetime):
    yield rf"\title{{{model.name} @ $d={dimension}$}}"
    if model.description:
        yield rf"\subtitle{{{model.description}}}"
    yield rf"\date{{{timestamp}}}"
    yield (
        rf"\author{{Generated by \texttt{{{autoeft.program}}}"
        rf" \textit{{{autoeft.version}}}}}"
    )


def tex_table(
    families: Count[str],
    types: Count[str],
    terms: Count[str],
    operators: Count[str],
) -> Iterator[str]:
    yield r"\begin{longtable}{crrr}"
    yield r"\toprule"
    yield r"family & types & terms & operators \\"
    yield r"\midrule"
    for family in families:
        yield (
            rf"${family}$ & {types[family]}"
            rf" & {terms[family]} & {operators[family]} \\"
        )
    yield r"\midrule"
    yield (
        rf"{sum(families.values())} & {sum(types.values())}"
        rf" & {sum(terms.values())} & {sum(operators.values())} \\"
    )
    yield r"\bottomrule"
    yield r"\end{longtable}"


def tex_series(
    hilbert_series: Count[str],
    dimension: int,
    add_hc: bool = False,
) -> Iterator[str]:
    yield r"\begin{align*}"
    yield r"\begin{autobreak}"
    yield rf"\hat{{H}}_{{{dimension}}} = "
    yield "  " + "\n+ ".join(
        rf"{co}\,{op}" if co > 1 else op for op, co in hilbert_series.items()
    )
    if add_hc:
        yield r"  + \hc"
    yield r"\end{autobreak}"
    yield r"\end{align*}"


def produce_texfile(path: Path, content: Iterable):
    disclaimer = f"% '{path.name}' generated by {autoeft.program} {autoeft.version}\n"
    path.write_text(disclaimer + "\n".join(content))



import itertools
from collections import Counter
from math import ceil
from math import floor
from typing import Counter as Count
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional

import sage.all
import sage.combinat.tableau

import autoeft.base.classifications as classes
import autoeft.base.groups as symmetry_groups
import autoeft.base.model as ir_model
import autoeft.combinatorics.symmetric_group as combinat
from autoeft import exceptions
from autoeft import utils
from autoeft.base import tensors


def modified_lr_rule(
    lhs: tensors.SUNTableau,
    rhs: tensors.SUNTableau,
    N: int,
) -> Iterator[tensors.SUNTableau]:
    """Modified Littlewood-Richardson rule.

    Return the product of two tableaux
    using a modified version of the Littlewood-Richardson rule.
    """
    for part, coeff in combinat.lr_rule(lhs.shape, rhs.shape, N).items():
        skew_tableaux = combinat.lr_skew_tableaux(part, lhs.shape, rhs.shape)
        assert len(skew_tableaux) == coeff
        for skew_tab in skew_tableaux:
            rhs_idx = [list(row) for row in rhs]
            yield tensors.SUNTableau(
                (
                    (
                        (
                            lhs[row, col]
                            if row < len(lhs) and col < len(lhs[row])
                            else rhs_idx[skew_tab[row, col] - 1].pop(0)
                        )
                        for col in range(row_len)
                    )
                    for row, row_len in enumerate(part)
                ),
            )


def multi_modified_lr_rule(
    tableaux: Iterable[tensors.SUNTableau],
    N: int,
    initials: Optional[Iterable[tensors.SUNTableau]] = None,
) -> Iterator[tensors.SUNTableau]:
    if initials is None:
        initials = [tensors.SUNTableau([])]
    it = iter(tableaux)
    try:
        rhs = next(it)
    except StopIteration:
        yield from initials
        return
    result = []
    for lhs in initials:
        result += list(modified_lr_rule(lhs, rhs, N))
    yield from multi_modified_lr_rule(it, N, result)


def construct_sun_tableaux(
    op_type: classes.Type,
    group: symmetry_groups.SUNGroup,
) -> List[tensors.SUNTableau]:
    """Return all tensors that transform in the invariant SU(N) representation.

    Raise 'NoContractionException' if no invariants are found.
    """
    partition = op_type.sun_partition(group)
    content = op_type.sun_content(group)
    if not partition and not any(content):
        return [tensors.SUNTableau()]
    tableaux = multi_modified_lr_rule(
        (
            tensors.SUNTableau.fill_tableau(field_repr, field_num)
            for field_num, field_repr in enumerate(content, start=1)
        ),
        group.N,
    )
    tableaux = sorted(tab for tab in tableaux if tab.shape == partition)
    if tableaux:
        return tableaux
    errmsg = group.name
    raise exceptions.NoContractionException(errmsg)  # no invariant contractions found


def construct_lorentz_tableaux(family: classes.Family) -> List[tensors.LorentzTableau]:
    """Return all Lorentz tableaux.

    Return all tensors for the given family that are Lorentz invariant
    and are not related to a total derivative.

    Raise 'NoContractionException' if no invariants are found.
    """
    partition = family.primary_partition()
    content = family.primary_content()
    if not partition and not any(content):
        return [tensors.LorentzTableau(0, 0)]
    assert sum(partition) == sum(content)
    SSYTx = sage.combinat.tableau.SemistandardTableaux
    tableaux = sorted(
        tensors.LorentzTableau(family.N, family.nr, tab)
        for tab in SSYTx(shape=partition, mu=content)
    )
    if tableaux:
        return tableaux
    errmsg = "Lorentz"
    raise exceptions.NoContractionException(errmsg)  # no invariant contractions found


def construct_types(
    family: classes.Family,
    model: ir_model.Model,
    hc_flag: bool = True,
) -> Iterator[classes.Type]:
    def _con_types(
        f_counts: Count[ir_model.Field],
        h_counts: Count[utils.Fraction],
        fields: List[ir_model.Field],
    ):
        if sum(h_counts.values()) == 0:  # all helicity slots are filled
            for name, group in model.symmetries.u1_groups.items():
                tot_charge = sum(
                    f.representations[name].charge * m for f, m in f_counts.items()
                )
                if abs(tot_charge - group.residual) > group.violation:
                    # reject operators with total charge greater than allowed violation
                    return
            for name, group in model.symmetries.sun_groups.items():
                tot_indices = sum(
                    sum(f.representations[name].partition) * m
                    for f, m in f_counts.items()
                )
                if tot_indices % group.N != 0:
                    # reject operators where the total number of SU(N) fundamental
                    # indices is not a multiple of N
                    return
            con_type = classes.Type(f_counts, family)
            if hc_flag or con_type.is_normal():
                yield con_type
        while fields:
            field = fields.pop(0)
            hel = field.helicity
            for nh in range(1, h_counts[hel] + 1):
                new_h_counts = h_counts.copy()
                new_h_counts[hel] -= nh
                new_f_counts = f_counts.copy()
                new_f_counts[field] += nh
                yield from _con_types(new_f_counts, new_h_counts, fields.copy())

    h_counts = Counter(
        dict(zip((utils.Fraction(h2, 2) for h2 in family.hel2), family.n_hel)),
    )
    yield from _con_types(Counter(), h_counts, list(model.fields.values()))


def _construct_families(
    N: int,
    nl: int,
    nr: int,
    hc_flag: bool = True,
) -> Iterator[classes.Family]:
    for nD in range(2 * min(nl, nr) + 1):
        for nFR in range(nr - ceil(nD / 2) + 1):
            for nFL in range(nl - ceil(nD / 2) + 1):
                nPsiD = 2 * nr - 2 * nFR - nD
                nPsi = 2 * nl - 2 * nFL - nD
                nPhi = N - nFR - nFL - nPsiD - nPsi
                n_hel = (nFL, nPsi, nPhi, nPsiD, nFR)
                fh2 = list(
                    itertools.chain.from_iterable(
                        itertools.repeat(h2, n)
                        for h2, n in zip((-2, -1, 0, 1, 2), n_hel)
                    ),
                )
                if (
                    all(nh >= 0 for nh in n_hel)
                    and nD >= max(nPsiD % 2, -nPsi % 2)
                    and nD >= -2 * min(fh2) - (2 * nFL + nPsi)
                    and nD >= 2 * max(fh2) - (2 * nFR + nPsiD)
                ):
                    con_family = classes.Family(n_hel, nD)
                    if hc_flag or con_family.is_normal():
                        yield con_family


def _construct_families_gr(
    N: int,
    nl: int,
    nr: int,
    nGR: int,
    hc_flag: bool = True,
) -> Iterator[classes.Family]:
    for nD in range(2 * min(nl, nr) + 1):
        for nCR in range(ceil(nr / 2 - nD / 4) + 1):
            nCL = nGR - nCR
            for nFR in range(nr - 2 * nCR - ceil(nD / 2) + 1):
                for nFL in range(nl - 2 * nCL - ceil(nD / 2) + 1):
                    nPsiD = 2 * nr - 4 * nCR - 2 * nFR - nD
                    nPsi = 2 * nl - 4 * nCL - 2 * nFL - nD
                    nPhi = N - nCR - nCL - nFR - nFL - nPsiD - nPsi
                    n_hel = (nCL, nFL, nPsi, nPhi, nPsiD, nFR, nCR)
                    fh2 = list(
                        itertools.chain.from_iterable(
                            itertools.repeat(h2, n)
                            for h2, n in zip((-4, -2, -1, 0, 1, 2, 4), n_hel)
                        ),
                    )
                    if (
                        all(nh >= 0 for nh in n_hel)
                        and nD >= max(nPsiD % 2, -nPsi % 2)
                        and nD >= -2 * min(fh2) - (4 * nCL + 2 * nFL + nPsi)
                        and nD >= 2 * max(fh2) - (4 * nCR + 2 * nFR + nPsiD)
                    ):
                        con_family = classes.FamilyGR(n_hel, nD)
                        if hc_flag or con_family.is_normal():
                            yield con_family


def construct_families(
    d: int,
    hc_flag: bool = True,
    gr_flag: bool = False,
) -> Iterator[classes.Family]:
    """Generate all families for the given dimension.

    The first elements corresponds to the so-called 'special kinematics',
    cf. arXiv:2005.00008 [hep-ph], p.16.
    """
    if not gr_flag:
        yield from _construct_families(3, d - 3, 0, hc_flag=hc_flag)
        if hc_flag:
            yield from _construct_families(3, 0, d - 3, hc_flag=hc_flag)
        for N in range(4, d + 1):
            bound_r = (d - N) if hc_flag else floor((d - N) / 2)
            for nr in range(bound_r + 1):
                yield from _construct_families(N, d - N - nr, nr, hc_flag=hc_flag)
    else:
        for nGR in range(floor(min(d / 2, 3)) + 1):
            yield from _construct_families_gr(3, d - 3 + nGR, 0, nGR, hc_flag=hc_flag)
            if hc_flag:
                yield from _construct_families_gr(
                    3,
                    0,
                    d - 3 + nGR,
                    nGR,
                    hc_flag=hc_flag,
                )
        for N in range(4, d + 1):
            for nGR in range(floor(min(d / 2, N)) + 1):
                bound_r = (d - N + nGR) if hc_flag else floor((d - N + nGR) / 2)
                for nr in range(bound_r + 1):
                    yield from _construct_families_gr(
                        N,
                        d - N - nr + nGR,
                        nr,
                        nGR,
                        hc_flag=hc_flag,
                    )



import os
import re
import shutil
import subprocess
from pathlib import Path
from typing import Any

import semver

from autoeft import exceptions

if (path := shutil.which("form", path=os.getenv("AUTOEFT_PATH"))) is None:
    errmsg = (
        "Unable to locate FORM."
        "\n"
        "Please make sure FORM is installed and added to the system PATH"
        " or to AUTOEFT_PATH."
    )
    raise exceptions.MissingRequirementError(errmsg)
executable = Path(path)
version_cmd = subprocess.run(
    (executable, "-version"),  # noqa: S603
    capture_output=True,
    text=True,
    check=True,
)
if (match := re.match(r"^FORM (\d+(?:\.\d+)*)", version_cmd.stdout)) is None:
    errmsg = "Invalid FORM version string '{version_cmd.stdout}'."
    raise exceptions.RequirementVersionError(errmsg)
__version__ = match[1]
version = semver.Version.parse(__version__, optional_minor_and_patch=True)


def form(obj: Any) -> str:
    """Call the FORM representation of an object."""
    return type(obj).__form__(obj)
